from django.urls import converters


class SignoffIdConverter(converters.SlugConverter):
    regex = "[-a-zA-Z0-9_.]+"  # a slug with the dot character


converters.register_converter(SignoffIdConverter, "id")


app_name = "signoffs"


urlpatterns = []



"""
Utility functions and classes
"""
import re

from django.core.exceptions import FieldDoesNotExist

split_caps_run = re.compile(r"(.)([A-Z][a-z]+)")
to_snake_case = re.compile(r"([a-z0-9])([A-Z])")
id_separators = re.compile(r"[_\-.]")


def camel_to_snake(name):
    """Convert CamelCaseName to snake_case_name"""
    # based on: https://stackoverflow.com/a/1176023/1993525
    name = split_caps_run.sub(r"\1_\2", name)
    return to_snake_case.sub(r"\1_\2", name).lower()


def id_to_camel(name):
    """Convert arbitrary identifier using dot or snake notation into CamelCase"""
    return "".join(el[:1].capitalize() + el[1:] for el in re.split(id_separators, name))


class Accessor(str):
    """
    A string describing a path from one object to another via attributes accesses.

    Relations are separated by a ``__`` character.

    Shamelessly adapted from the amazing django_tables2 https://pypi.org/project/django-tables2/
    """

    SEPARATOR = "__"

    ALTERS_DATA_ERROR_FMT = "Refusing to call {method}() because `.alters_data = True`"
    LOOKUP_ERROR_FMT = "Failed lookup for attribute [{attr}] in {obj}, when resolving the accessor `{accessor}`"

    def resolve(self, obj, safe=True, quiet=False):
        """
        Return an attribute described by the accessor by traversing the attributes of object

        Callable objects are called, and their result is used, before proceeding with the resolving.
        Usage:
        ```
            >>> from django.contrib.auth.models import User
            >>> user = User(first_name='Brad')
            >>> x = Accessor("user__first_name")
            >>> x.resolve(user)
            "Brad"
        ```

        :param obj: The root object to traverse.
        :param bool safe: Don't call anything with `alters_data = True`
        :param bool quiet: Smother all exceptions and instead return `None`
        :return: resolved target object

        :raises TypeError, AttributeError, KeyError, ValueError: (unless `quiet` == `True`)
        """

        def traverse(current, bit):
            """Traverse to current.bit and return the result or raise ValueError if no such relation"""
            try:
                return getattr(current, bit)
            except AttributeError:
                raise ValueError(
                    self.LOOKUP_ERROR_FMT.format(attr=bit, obj=current, accessor=self)
                )

        def check_safe(item):
            """Raise ValueError if item is callable and item.alters_data but safe==True"""
            if callable(current) and safe and getattr(current, "alters_data", False):
                raise ValueError(
                    self.ALTERS_DATA_ERROR_FMT.format(method=repr(current))
                )

        # Short-circuit if the object has an attribute with the exact name of the accessor,
        try:
            return None if self == "" else getattr(obj, self)
        except AttributeError:
            try:
                current = obj
                for bit in self.bits:
                    current = traverse(current, bit)
                    check_safe(current)
                    # Important that we break in None case; otherwise a relationship spanning
                    #  a null-key will raise an exception in the next iteration, instead of defaulting.
                    if current is None:
                        break
                return current
            except Exception:
                if not quiet:
                    raise

    @property
    def bits(self):
        return self.split(self.SEPARATOR) if self != "" else ()

    def get_field(self, model):
        """
        Return the django model field for model in context, following relations.
        """
        if not hasattr(model, "_meta"):
            return

        field = None
        for bit in self.bits:
            try:
                field = model._meta.get_field(bit)
            except FieldDoesNotExist:
                break

            if hasattr(field, "remote_field"):
                rel = getattr(field, "remote_field", None)
                model = getattr(rel, "model", model)

        return field

    def penultimate_accessor(self):
        """
        Split off the right-most separator.

        :return: a 2-tuple (the Accessor for the left part, the remainder right-part)

        Usage:
        ```
            >>> Accessor("user__profile__title").penultimate_accessor()
            'user__profile', 'title'
        ```
        """
        path, _, remainder = self.rpartition(self.SEPARATOR)
        return Accessor(path), remainder

    def penultimate(self, obj, quiet=True):
        """
        Split the accessor on the right-most separator ('__'),

        :return: a 2-tuple (the resolved left part, the remainder right-part)

        Usage:
        ```
            >>> Accessor("user__profile__title").penultimate(user)
            <Profile object>, 'title'
        ```
        """
        accessor, remainder = self.penultimate_accessor()
        return accessor.resolve(obj, quiet=quiet), remainder


class ServiceDescriptor:
    """
    A descriptor used to "inject" instances of a "service" class into its owner's instances.

    A "service" provides services or strategies to its owner, but needs the owner instance to do its own work.
    Construction of the owner instance may not be under direct control, so service instantiation must be automated.
    Service class must expect owner instance as first positional parameter of its constructor.
    """

    service_class = None

    def __init__(self, service_class=None, **kwargs):
        """
        Inject `service_class` instance into the instance of the descriptor's owner class

        first positional arg for `service_class` initializer must be an instance of owner class
        `kwargs` are passed through to the `service_class` initializer
        """
        self.service_class = service_class or self.service_class
        self.service_class_kwargs = kwargs
        self.attr_name = ""  # set by __set_name__

    def __set_name__(self, owner, name):
        self.attr_name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self.service_class
        else:
            service_obj = self.service_class(instance, **self.service_class_kwargs)
            setattr(instance, self.attr_name, service_obj)
            return service_obj


def service(service_class, **kwargs):
    """
    Factory to return specialized service descriptors.

    :return: a `ServiceDescriptor` for a specialized subclass of `service_class`, that has `kwargs` as class attributes

    Usage:
    ```
        >>> class Service:
        ...     service_type = 'generic'
        ...     def __init__(self, owner, extra=None):
        ...         self.owner = owner
        ...         self.extra = extra
        ...     def __str__(self):
        ...         extra = f' with {self.extra}' if self.extra else ''
        ...         return f'A {self.service_type} service for {self.owner}{extra}'
        ... class Owner:
        ...     a_service = service(Service, service_type='special')(extra="whazoo")
        ...     def __str__(self):
        ...         return 'Owner'
        ... o = Owner()
        ... assert str(o.a_service) == "A special service for Owner with whazoo"
    ```
    """
    specialized_service = type(service_class.__name__, (service_class,), kwargs)

    descriptor_name = f"{service_class.__name__}Service"
    descriptor = type(
        descriptor_name, (ServiceDescriptor,), dict(service_class=specialized_service)
    )
    return descriptor


class ClassServiceDescriptor:
    """
    A descriptor used to "inject" instances of a "service" class onto its owner class.

    This is analogous to `ServiceDescriptor` but service instance is available on owner class
    First positional parameter of `service_class` class must be an owner class (type not instance!)
    """

    service_class = None

    def __init__(self, service_class=None, **kwargs):
        """
        Inject `service_class` instance, initialized with owner class, into the descriptor's owner class

        first positional arg for `service_class` initializer must be an owner class type
        `kwargs` are passed through to the `service_class` initializer
        """
        self.service_class = service_class or self.service_class
        self.service_class_kwargs = kwargs
        self.attr_name = ""  # set by __set_name__

    def __set_name__(self, owner, name):
        self.attr_name = name

    def __get__(self, instance, owner):
        service_obj = self.service_class(owner, **self.service_class_kwargs)
        setattr(owner, self.attr_name, service_obj)
        return service_obj


def class_service(service_class, **kwargs):
    """
    Factory to return specialized class service descriptors.

    :return: a `ClassServiceDescriptor` for a specialized subclass of `service_class`,
    that has `kwargs` as class attributes
    """
    specialized_service = type(service_class.__name__, (service_class,), kwargs)

    descriptor_name = f"{service_class.__name__}ClassService"
    descriptor = type(
        descriptor_name,
        (ClassServiceDescriptor,),
        dict(service_class=specialized_service),
    )
    return descriptor


__all__ = [
    "service",
    "ServiceDescriptor",
    "class_service",
    "ClassServiceDescriptor",
]



"""
A "Stamp of Approval" is the persistence layer that backs an Approval - the timestamp that seals the Approval.
    - often models the set of related Signets that define the Approval's "Signing Order".
A concrete Stamp may have other information about, or perhaps a relation to, the thing being approved.

A Stamp records status of an Approval, and is often automated by business logic defining when the Approval is complete.
An approved Stamp record provides evidence that a particular Approval was granted and by whom (via signatories).
A Stamp may be granted directly if business rules dictate, but normally is granted via one or more Signets.
Otherwise, a Stamp is not ordinarily edited directly - add and revoke Signets to move it through its signing order.
Once granted, the Stamp persists its "approved" status, regardless of the state of underlying Signets or
    changes to the Approval process.
To revoke a Stamp, we alter the approval status and revoke the Signet(s) used to grant the Approval.
A "blame" history, may be maintained by using a RevokeSignet model on the Approval Type.
"""
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.utils import timezone

from .signets import AbstractSignet


class AbstractApprovalSignet(AbstractSignet):
    """A Signet representing one signature on an Approval. The Approval Stamp related_name must be "signatories" """

    stamp = models.ForeignKey(
        "Stamp", on_delete=models.CASCADE, related_name="signatories"
    )

    class Meta(AbstractSignet.Meta):
        abstract = True


class ApprovalStampQuerySet(models.QuerySet):
    """
    Custom queries for Approval Seal
    """

    def approved(self):
        """Filter out unapproved stamps"""
        return self.filter(approved=True)

    def incomplete(self):
        """Return only stamps that are not yet approved"""
        return self.exclude(approved=True)

    def prefetch_signets(self):
        """Prefetch related signets but not the signing users"""
        return self.prefetch_related("signatories")

    def prefetch_signatories(self):
        """Prefetch related signets and their signing users"""
        return self.prefetch_related("signatories__user")

    def approvals(self, approval_id=None, subject=None):
        """
        Returns list of approval objects, one for each seal in queryset,
            optionally filtered for specific approval type - filtering done in-memory for performance.
        """
        return [
            seal.get_approval(subject=subject)
            for seal in self
            if approval_id is None or seal.approval_id == approval_id
        ]


ApprovalStampManager = models.Manager.from_queryset(ApprovalStampQuerySet)


def validate_approval_id(value):
    """Raise ValidationError if value is not a registered Approval Type ID"""
    from signoffs import registry

    if value is None or value not in registry.approvals:
        raise ValidationError(f"Invalid or unregistered approval {value}")


class AbstractApprovalStamp(models.Model):
    """
    Abstract base class for all "Stamp of Approval" models
    A Stamp is the model often on the One side of a Many-To-One relation from an ApprovalSignet
    Persistence layer that timestamps an approval: who, when, what
        (what is supplied by concrete class, e.g., with a FK to other model)
    """

    approval_id = models.CharField(
        max_length=100,
        null=False,
        validators=[validate_approval_id],
        verbose_name="Approval Type",
    )
    approved = models.BooleanField(default=False, verbose_name="Approved")
    # timestamp the approval - updated by approve() method
    timestamp = models.DateTimeField(default=timezone.now, editable=False, null=False)

    class Meta:
        abstract = True
        ordering = [
            "timestamp",
        ]  # chronological order

    objects = ApprovalStampManager()

    read_only_fields = (
        "approval_id",
        "timestamp",
    )  # fields managed in code cannot be manipulated

    def __str__(self):
        return (
            f"{self.approval_id} at {self.timestamp}"
            if self.is_approved()
            else f"{self.approval_id} (incomplete)"
        )

    @property
    def approval_type(self):
        """Return the Approval Type (class) that governs this stamp"""
        from signoffs.registry import approvals

        approval_type = approvals.get(self.approval_id)
        if not approval_type:
            raise ImproperlyConfigured(
                f"""Approval type {approval_type} not registered.
            See AUTODISCOVER settings to discover approval types when django loads."""
            )
        return approval_type

    def get_approval(self, subject=None):
        """Return an Approval instance for this stamp"""
        return self.approval_type(stamp=self, subject=subject)

    @property
    def approval(self):
        """The Approval instance for this stamp"""
        return self.get_approval()

    def is_approved(self):
        """return True if this Stamp is approved and has a persistent representation in DB"""
        return self.approved and self.id is not None

    def approve(self):
        """
        Approve the stamp (but don't commit the change)
        No permissions involved here - just force this stamp into approved state!
        """
        self.approved = True
        self.timestamp = timezone.now()

    def is_user_signatory(self, user):
        """return True iff the given user is a signatory on this stamp"""
        return any(s.user == user for s in self.signatories.all())

    def has_valid_approval(self):
        """return True iff this Stamp has a valid approval_id"""
        from signoffs import registry

        return self.approval_id is not None and self.approval_id in registry.approvals

    def save(self, *args, **kwargs):
        """Validate and save this stamp"""
        self.full_clean()
        return super().save(*args, **kwargs)

    @classmethod
    def has_object_relation(cls):
        """Return True iff this Stamp class has a FK relation to some object other than user"""
        relations = [
            f
            for f in cls._meta.fields
            if isinstance(f, models.ForeignKey)
            and not issubclass(f.related_model, AbstractApprovalStamp)
            and f.name != "user"
        ]
        return bool(relations)



"""
    Views for demonstrating and testing templates, templatetags, and rendering logic

    TODO: extend these views so they can be used to drive an approval process.
        add post form handling logic, etc.
"""
from django.views import generic

from signoffs.shortcuts import get_approval_stamp_or_404, get_signet_or_404


class SignoffDetailView(generic.DetailView):
    template_name = "test_app/signets/detail_view.html"

    def get_object(self, queryset=None):
        """Return the signet instance for the given signoff type"""
        return get_signet_or_404(self.kwargs["signoff_id"], self.kwargs["pk"])

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            signed_signoff=self.object.signoff,
            unsigned_signoff=type(self.object.signoff)(),
            **kwargs,
        )


class ApprovalDetailView(generic.DetailView):
    template_name = "test_app/approvals/detail_view.html"

    def get_object(self, queryset=None):
        """Return the signet instance for the given signoff type"""
        return get_approval_stamp_or_404(self.kwargs["approval_id"], self.kwargs["pk"])

    def get_context_data(self, **kwargs):
        return super().get_context_data(approval=self.object.approval, **kwargs)



"""
Test App - tests for approval process models
"""
from django.test import TestCase

from signoffs.core.tests import fixtures
from tests.test_app.models import Building, ConstructionPermittingProcess


class FsmApprovalProcessModelTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.applicant = fixtures.get_user()
        cls.planner = fixtures.get_user()
        cls.electrician = fixtures.get_user()
        cls.plumber = fixtures.get_user()
        cls.inspector = fixtures.get_user()

    def test_fsm_transition_sequence(self):
        model = ConstructionPermittingProcess(
            building=Building.objects.create(name="Eiffel Tower")
        )
        self.assertListEqual(
            [model.applied, model.permitted, model.inspected, model.authorized],
            [
                model.process.bound_approve_transition(a)
                for a in model.process.get_all_approvals()
            ],
        )

    def test_fsm_get_all_approvals(self):
        model = ConstructionPermittingProcess(
            building=Building.objects.create(name="Eiffel Tower")
        )
        self.assertListEqual(
            model.process.get_all_approvals(),
            [
                model.apply,
                model.permit,
                model.interim_inspection,
                model.final_inspection,
            ],
        )

    def test_fsm_get_available_approvals(self):
        model = ConstructionPermittingProcess(
            building=Building.objects.create(name="Hotel California")
        )
        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.apply,
            ],
        )
        self.permit_application(model)
        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.permit,
            ],
        )
        self.permit_approval(model)
        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.interim_inspection,
            ],
        )
        self.interim_inspection(model)
        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.final_inspection,
            ],
        )
        self.final_inspection(model)
        self.assertListEqual(model.process.get_available_approvals(), [])

    # Approval Steps for ConstructionPermittingProcess

    def permit_application(self, model):
        """move the given process from INITIATED to APPLIED"""
        self.assertEqual(model.state, model.States.INITIATED)
        self.assertTrue(model.process.can_proceed(model.apply))
        self.assertFalse(model.process.can_proceed(model.interim_inspection))
        self.assertTrue(
            model.process.user_can_proceed(user=self.applicant, approval=model.apply)
        )

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.apply)

        next_approval.sign_approval(user=self.applicant)
        self.assertTrue(
            model.process.try_approve_transition(model.apply, user=self.applicant)
        )
        self.assertTrue(model.apply.is_approved())
        self.assertEqual(model.state, model.States.APPLIED)

    def revoke_application(self, model):
        """move the given process from APPLIED back to INITIATED"""
        self.assertEqual(model.state, model.States.APPLIED)
        self.assertTrue(model.process.is_revokable(model.apply))
        self.assertFalse(model.process.is_revokable(model.interim_inspection))
        self.assertTrue(
            model.process.user_can_revoke(user=self.applicant, approval=model.apply)
        )

        revoke_approval = model.process.get_next_revokable_approval()
        self.assertEqual(revoke_approval, model.apply)

        self.assertTrue(
            model.process.try_revoke_transition(model.apply, user=self.applicant)
        )
        self.assertFalse(model.apply.is_approved())
        self.assertEqual(model.state, model.States.INITIATED)

    def permit_approval(self, model):
        """move the given process from APPLIED to PERMITTED"""
        self.assertEqual(model.state, model.States.APPLIED)
        self.assertTrue(model.process.can_proceed(model.permit))
        self.assertFalse(model.process.can_proceed(model.interim_inspection))

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.permit)

        next_approval.sign_approval(user=self.planner)
        self.assertFalse(model.permit.is_approved())
        self.assertEqual(model.state, model.States.APPLIED)

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.permit)

        next_approval.sign_approval(user=self.electrician)
        self.assertFalse(model.permit.is_approved())
        self.assertEqual(model.state, model.States.APPLIED)

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.permit)

        next_approval.sign_approval(user=self.inspector)
        self.assertTrue(
            model.process.try_approve_transition(model.permit, user=self.inspector)
        )
        self.assertTrue(model.permit.is_approved())
        self.assertEqual(model.state, model.States.PERMITTED)

    def interim_inspection(self, model):
        """move the given process from PERMITTED to INSPECTED"""
        self.assertEqual(model.state, model.States.PERMITTED)
        self.assertTrue(model.process.can_proceed(model.interim_inspection))
        self.assertFalse(model.process.can_proceed(model.final_inspection))

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.interim_inspection)

        next_approval.sign_approval(user=self.electrician)
        self.assertFalse(model.interim_inspection.is_approved())
        self.assertEqual(model.state, model.States.PERMITTED)

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.interim_inspection)

        next_approval.sign_approval(
            user=self.plumber, last=True
        )  # TODO: remove with permissions or transition restrictions
        self.assertFalse(model.interim_inspection.is_approved())
        self.assertEqual(model.state, model.States.PERMITTED)

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.interim_inspection)

        next_approval.sign_approval(
            user=self.inspector, last=True
        )  # TODO: remove with permissions or transition restrictions
        self.assertTrue(
            model.process.try_approve_transition(
                model.interim_inspection, user=self.inspector
            )
        )
        self.assertTrue(model.interim_inspection.is_approved())
        self.assertEqual(model.state, model.States.INSPECTED)

    def final_inspection(self, model):
        """move the given process from INSPECTED to APPROVED"""
        self.assertEqual(model.state, model.States.INSPECTED)
        self.assertTrue(model.process.can_proceed(model.final_inspection))
        self.assertFalse(model.process.can_proceed(model.apply))
        self.assertFalse(model.process.can_proceed(model.permit))
        self.assertFalse(model.process.can_proceed(model.interim_inspection))

        next_approval = model.process.get_next_available_approval()
        self.assertEqual(next_approval, model.final_inspection)

        next_approval.sign_approval(
            user=self.inspector, last=True
        )  # TODO: remove with permissions or transition restrictions
        self.assertTrue(
            model.process.try_approve_transition(
                model.final_inspection, user=self.inspector
            )
        )
        self.assertTrue(model.final_inspection.is_approved())
        self.assertEqual(model.state, model.States.APPROVED)

    def test_fsm_approval_sequencing(self):
        model = ConstructionPermittingProcess(
            building=Building.objects.create(name="Burj Khalifa")
        )
        self.permit_application(model)
        self.permit_approval(model)
        self.interim_inspection(model)
        self.final_inspection(model)

    def test_fsm_revoke_transition(self):
        model = ConstructionPermittingProcess(
            building=Building.objects.create(name="Burj Khalifa")
        )
        self.permit_application(model)
        self.revoke_application(model)
        self.permit_application(model)
        self.revoke_application(model)



"""
    Proxy for Approval Process to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""

from signoffs.core.process import (
    ApprovalsProcess,
    FsmApprovalsProcess,
    user_can_revoke_approval,
)



"""
Concrete models, signoffs, approvals, etc. used by  test suite
"""
import django.forms
from django.db import models

import signoffs.core.signing_order as so
from signoffs.core.approvals import BaseApproval
from signoffs.core.forms import AbstractSignoffForm, SignoffFormsManager
from signoffs.core.models import (
    AbstractApprovalSignet,
    AbstractApprovalStamp,
    AbstractRevokedSignet,
    AbstractSignet,
)
from signoffs.core.models.fields import ApprovalField, SignoffField, SignoffSet
from signoffs.core.signoffs import BaseSignoff
from signoffs.registry import register

# Concrete Signet Models


class Signet(AbstractSignet):
    pass


class OtherSignet(AbstractSignet):
    pass


class LeaveSignet(AbstractSignet):
    object = models.ForeignKey(
        "signoffs.LeaveRequest", on_delete=models.CASCADE, related_name="signatories"
    )


class RevokedLeaveSignet(AbstractRevokedSignet):
    signet = models.OneToOneField(
        LeaveSignet, on_delete=models.CASCADE, related_name="revoked"
    )


class ApprovalSignet(AbstractApprovalSignet):
    pass


# Signoffs backed by the Signet models above


class BasicSignoff(BaseSignoff):
    signetModel = Signet
    label = "Consent?"


simple_signoff_type = BasicSignoff.register(id="test.signoffs.simple_signoff")


def approval_signoff_form():
    class ApprovalSignoffForm(AbstractSignoffForm):
        """Form for collecting approval signoffs"""

        class Meta(AbstractSignoffForm.Meta):
            model = ApprovalSignet
            widgets = {"stamp": django.forms.HiddenInput}

    return ApprovalSignoffForm


class ApprovalSignoff(BaseSignoff):
    signetModel = ApprovalSignet
    forms = SignoffFormsManager(signoff_form=approval_signoff_form)


class LeaveSignoff(BaseSignoff):
    signetModel = "signoffs.LeaveSignet"
    revokeModel = "signoffs.RevokedLeaveSignet"
    label = "Consent?"


class InvalidModel(models.Model):
    invalid_signet = SignoffSet("test.signoffs.simple_signoff")
    invalid_relation = SignoffSet("test.approval.leave.employee_signoff")


# Concrete Stamp models


class Stamp(AbstractApprovalStamp):
    pass


class OtherStamp(AbstractApprovalStamp):
    pass


# A LeaveRequest model with a LeaveApproval process defined in 2 different ways


class AbstractLeaveApproval(BaseApproval):
    stampModel = Stamp


@register(id="test.approval.fields.leave_approval")
class LeaveApproval(AbstractLeaveApproval):
    label = "Approve Leave of Absence"

    employee_signoff_type = ApprovalSignoff.register(
        id="test.approval.leave.employee_signoff"
    )
    hr_signoff_type = ApprovalSignoff.register(id="test.approval.leave.hr_signoff")
    mngmt_signoff_type = ApprovalSignoff.register(
        id="test.approval.leave.mngmt_signoff"
    )

    signing_order = so.SigningOrder(
        employee_signoff_type, so.AtLeastN(hr_signoff_type, n=1), mngmt_signoff_type
    )


class LeaveRequest(models.Model):
    """Demonstrates 2 ways to manage a group of signoffs on a model (wouldn't normally use both!!)"""

    # (1) directly on the Model class (signing order is managed by model business logic) (e.g., see signoff tests)
    employee_signoff_type = BasicSignoff.register(id="test.leave.employee_signoff")
    hr_signoff_type = LeaveSignoff.register(id="test.leave.hr_signoff")
    mngmt_signoff_type = LeaveSignoff.register(id="test.leave.mngmt_signoff")

    # One-to-One "forward" relation - the OneToOneField is named employee_signet
    employee_signoff, employee_signet = SignoffField(employee_signoff_type)
    # One-to-Many "reverse" relation based on relation defined by LeaveSignet
    hr_signoffs = SignoffSet(hr_signoff_type)
    mngmt_signoffs = SignoffSet(mngmt_signoff_type)

    # (2) a One-to-One "forward" approval relation (managed by the Approval.signing_order) (e.g., see approval tests)
    approval, approval_stamp = ApprovalField(LeaveApproval)



"""
App-independent Approval Process INTEGRATION  tests
"""
from django.db import models
from django.test import TestCase
from django_fsm import FSMField, transition

import signoffs.core.signing_order as so
from signoffs.core import process as signoffs_process
from signoffs.core.models.fields import ApprovalField, ApprovalSignoffSingle
from signoffs.registry import register

from . import fixtures
from .models import AbstractLeaveApproval, ApprovalSignoff


@register(id="test.process_approval.leave_application")
class LeaveApplication(AbstractLeaveApproval):
    label = "Apply for Leave of Absence"

    employee_signoff = ApprovalSignoffSingle(
        ApprovalSignoff.register(id="test.process_approval.leave.employee_signoff")
    )

    def is_complete(self):
        """Customize logic for determining if Approval signoffs are complete and ready to be approved."""
        return self.employee_signoff.get().is_signed()

    def sign_application(self, user):
        if self.employee_signoff.can_sign(user):
            self.employee_signoff.create(user=user)


class ApprovalSignoffSingleTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.employee = fixtures.get_user(first_name="Employee")

    def test_signoff_single(self):
        approval = LeaveApplication.create()
        self.assertFalse(approval.employee_signoff.exists())
        self.assertTrue(approval.employee_signoff.can_sign(self.employee))
        approval.sign_application(user=self.employee)
        self.assertTrue(approval.employee_signoff.exists())
        self.assertFalse(approval.employee_signoff.can_sign(self.employee))
        # None oof these do anything
        approval.sign_application(user=self.employee)
        approval.sign_application(user=self.employee)
        self.assertFalse(approval.employee_signoff.can_sign(self.employee))
        self.assertEqual(approval.employee_signoff.count(), 1)


@register(id="test.process_approval.leave_approval")
class LeaveApproval(AbstractLeaveApproval):
    label = "Approve Leave of Absence"

    hr_signoff_type = ApprovalSignoff.register(
        id="test.process_approval.leave.hr_signoff"
    )
    mngmt_signoff_type = ApprovalSignoff.register(
        id="test.process_approval.leave.mngmt_signoff"
    )

    signing_order = so.SigningOrder(
        so.AtLeastN(hr_signoff_type, n=1), mngmt_signoff_type
    )

    def sign_approval(self, user):
        next = self.next_signoffs(for_user=user)
        if next and not self.is_approved() and not self.has_signed(user):
            next[-1].sign(
                user
            )  # Arbitrarily approve "most nextest" signoff on the approval


#  SIMPLE TEST CASES


class SimpleApprovalProcessModel(models.Model):
    """An Approval Process model where transitions and approval sequencing are handled by app logic - no ordering"""

    class States(models.TextChoices):
        INITIATED = "Initiated"
        APPLIED = "Applied"
        APPROVED = "Approved"

    state = "Initiated"

    apply, apply_stamp = ApprovalField(LeaveApplication)
    approve, approve_stamp = ApprovalField(LeaveApproval)

    process = (
        signoffs_process.ApprovalsProcess()
    )  # optionally list approvals in order, default is order they are registered below

    @process.register_and_do_approval(apply)
    def application_made(self, approval):
        self.state = self.States.APPLIED
        print("Application made!", self.state, self.apply)

    @process.register_and_do_approval(approve)
    def leave_approved(self, approval):
        self.state = self.States.APPROVED
        print("Leave is approved!", self.state, self.approve)


class SimpleApprovalProcessModelTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.employee = fixtures.get_user(first_name="Employee")
        cls.hr = fixtures.get_user(first_name="HR")
        cls.mngr = fixtures.get_user(first_name="Manager")

    def test_approve_transitions_sequence(self):
        model = SimpleApprovalProcessModel()
        self.assertListEqual(
            [model.application_made, model.leave_approved],
            [
                model.process.bound_approve_transition(a)
                for a in model.process.get_all_approvals()
            ],
        )

    def test_get_all_approvals(self):
        model = SimpleApprovalProcessModel()
        self.assertListEqual(
            model.process.get_all_approvals(), [model.apply, model.approve]
        )

    def test_get_approved_approvals(self):
        model = SimpleApprovalProcessModel()
        self.assertListEqual(model.process.get_approved_approvals(), [])
        self.assertListEqual(
            model.process.get_unapproved_approvals(), [model.apply, model.approve]
        )
        model.apply.approve()
        self.assertListEqual(model.process.get_approved_approvals(), [model.apply])
        self.assertListEqual(model.process.get_unapproved_approvals(), [model.approve])
        model.approve.approve()
        self.assertListEqual(
            model.process.get_approved_approvals(), [model.apply, model.approve]
        )
        self.assertListEqual(model.process.get_unapproved_approvals(), [])

    def test_get_available_approvals(self):
        model = SimpleApprovalProcessModel()
        self.assertListEqual(model.process.get_available_approvals(), [model.apply])
        model.apply.approve()
        self.assertListEqual(model.process.get_available_approvals(), [model.approve])
        model.approve.approve()
        self.assertListEqual(model.process.get_available_approvals(), [])

    def test_approval_actions(self):
        model = SimpleApprovalProcessModel()
        self.assertEqual(model.state, model.States.INITIATED)
        model.apply.sign_application(user=self.employee)
        self.assertTrue(
            model.process.try_approve_transition(model.apply, self.employee)
        )
        self.assertEqual(model.state, model.States.APPLIED)
        model.approve.sign_approval(user=self.hr)
        self.assertFalse(model.process.try_approve_transition(model.approve, self.hr))
        self.assertEqual(model.state, model.States.APPLIED)
        model.approve.sign_approval(user=self.mngr)
        self.assertTrue(model.process.try_approve_transition(model.approve, self.mngr))
        self.assertEqual(model.state, model.States.APPROVED)


#  ORDERED TEST CASES


class OrderedApprovalProcessModel(models.Model):
    """An Approval Process model where transitions and approval sequencing are defined with linear ordering"""

    class States(models.TextChoices):
        INITIATED = "Initiated"
        APPLIED = "Applied"
        APPROVED = "Approved"

    state = "Initiated"

    zapply, zapply_stamp = ApprovalField(LeaveApplication)
    approve, approve_stamp = ApprovalField(LeaveApproval)

    process = signoffs_process.ApprovalsProcess(
        zapply, approve
    )  # list approvals in order so code order doesn't matter

    @process.register_and_do_approval(approve)
    def leave_approved(self, approval):
        self.state = self.States.APPROVED
        print("Leave is approved!", self.state, self.approve)

    @process.register_and_do_approval(zapply)
    def application_made(self, approval):
        self.state = self.States.APPLIED
        print("Application made!", self.state, self.zapply)


class OrderedApprovalProcessModelTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.employee = fixtures.get_user(first_name="Employee")
        cls.hr = fixtures.get_user(first_name="HR")
        cls.mngr = fixtures.get_user(first_name="Manager")

    def test_approval_sequence(self):
        model = OrderedApprovalProcessModel()
        self.assertListEqual(
            [model.application_made, model.leave_approved],
            [
                model.process.bound_approve_transition(a)
                for a in model.process.get_all_approvals()
            ],
        )

    def test_get_all_approvals(self):
        model = OrderedApprovalProcessModel()
        self.assertListEqual(
            model.process.get_all_approvals(), [model.zapply, model.approve]
        )

    def test_approval_ordering(self):
        model = OrderedApprovalProcessModel()
        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.zapply,
            ],
        )
        self.assertEqual(model.process.get_next_available_approval(), model.zapply)
        self.assertTrue(model.process.can_proceed(model.zapply))
        self.assertFalse(model.process.can_proceed(model.approve))
        model.zapply.approve()

        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.approve,
            ],
        )
        self.assertEqual(model.process.get_next_available_approval(), model.approve)
        self.assertFalse(model.process.can_proceed(model.zapply))
        self.assertTrue(model.process.can_proceed(model.approve))
        model.approve.approve()

        self.assertListEqual(model.process.get_available_approvals(), [])
        self.assertEqual(model.process.get_next_available_approval(), None)
        self.assertFalse(model.process.can_proceed(model.zapply))
        self.assertFalse(model.process.can_proceed(model.approve))

    def test_approval_actions(self):
        model = OrderedApprovalProcessModel()
        self.assertEqual(model.state, model.States.INITIATED)
        model.zapply.sign_application(user=self.employee)
        self.assertTrue(
            model.process.try_approve_transition(model.zapply, self.employee)
        )
        self.assertEqual(model.state, model.States.APPLIED)
        model.approve.sign_approval(user=self.hr)
        self.assertFalse(model.process.try_approve_transition(model.approve, self.hr))
        self.assertEqual(model.state, model.States.APPLIED)
        model.approve.sign_approval(user=self.mngr)
        self.assertTrue(model.process.try_approve_transition(model.approve, self.mngr))
        self.assertEqual(model.state, model.States.APPROVED)


# FSM Test Cases  TODO: add revoke transitions, transition conditions & permissions


class FsmLeaveApprovalProcessModel(models.Model):
    """An Approval Process Model where state transitions and approval sequencing are ordered using fsm"""

    class States(models.TextChoices):
        INITIATED = "Initiated"
        APPLIED = "Applied"
        APPROVED = "Approved"

    state = FSMField(choices=States.choices, default=States.INITIATED)

    apply, apply_stamp = ApprovalField(LeaveApplication)
    approve, approve_stamp = ApprovalField(LeaveApproval)

    process = (
        signoffs_process.FsmApprovalsProcess()
    )  # adds FSM transition logic to approvals process

    # Decorate transition the long (but flexible) way...  (* note: decorateor order really matters here!)

    @process.register_approve_transition(apply)
    @transition(field=state, source=States.INITIATED, target=States.APPLIED)
    @process.do_approval
    def application_made(self, approval):
        print("Application made!", self.state, self.apply)

    # or the convenient way

    @process.approval_transition(
        approve, field=state, source=States.APPLIED, target=States.APPROVED
    )
    def leave_approved(self, approval):
        print("Leave is approved!", self.state, self.approve)


class FsmApprovalProcessModelTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.employee = fixtures.get_user(first_name="Employee")
        cls.hr = fixtures.get_user(first_name="HR")
        cls.mngr = fixtures.get_user(first_name="Manager")

    def test_fsm_approval_sequence(self):
        model = FsmLeaveApprovalProcessModel()
        self.assertListEqual(
            [model.application_made, model.leave_approved],
            [
                model.process.bound_approve_transition(a)
                for a in model.process.get_all_approvals()
            ],
        )

    def test_fsm_get_available_approvals(self):
        model = FsmLeaveApprovalProcessModel()
        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.apply,
            ],
        )
        model.apply.sign_application(user=self.employee)
        model.process.try_approve_transition(model.apply, user=self.employee)

        self.assertListEqual(
            model.process.get_available_approvals(),
            [
                model.approve,
            ],
        )
        model.approve.sign_approval(user=self.hr)
        model.approve.sign_approval(user=self.mngr)
        model.process.try_approve_transition(model.approve, user=self.mngr)
        self.assertListEqual(model.process.get_available_approvals(), [])

    def test_fsm_approval_sequencing(self):
        model = FsmLeaveApprovalProcessModel()
        self.assertEqual(model.state, model.States.INITIATED)
        self.assertTrue(model.process.can_proceed(model.apply))
        self.assertFalse(model.process.can_proceed(model.approve))

        model.apply.sign_application(user=self.employee)
        self.assertTrue(
            model.process.try_approve_transition(model.apply, user=self.employee)
        )
        self.assertEqual(model.state, model.States.APPLIED)
        self.assertFalse(model.process.can_proceed(model.apply))
        self.assertTrue(model.process.can_proceed(model.approve))

        model.approve.sign_approval(user=self.hr)
        self.assertFalse(
            model.process.try_approve_transition(model.approve, user=self.hr)
        )
        self.assertEqual(model.state, model.States.APPLIED)
        self.assertFalse(model.process.can_proceed(model.apply))
        self.assertTrue(model.process.can_proceed(model.approve))

        model.approve.sign_approval(user=self.hr)
        self.assertFalse(
            model.process.try_approve_transition(model.approve, user=self.hr)
        )
        self.assertEqual(model.state, model.States.APPLIED)
        self.assertFalse(model.process.can_proceed(model.apply))
        self.assertTrue(model.process.can_proceed(model.approve))

        model.approve.sign_approval(user=self.mngr)
        self.assertTrue(
            model.process.try_approve_transition(model.approve, user=self.mngr)
        )
        self.assertEqual(model.state, model.States.APPROVED)
        self.assertFalse(model.process.can_proceed(model.apply))
        self.assertFalse(model.process.can_proceed(model.approve))



"""
    A Signoff defines the business and presentation logic for collecting a single "signature"

    Signoff Types are registered subclasses of AbstractSignoff
        - they define the behaviour for a Signoff.
    Persistence layer for Signoff state is provided by a Signet model
        - one concrete Signet model can back any number of Signoff Types
        - can think of a Signoff instance as the strategy for managing a Signet instance.
"""
from typing import Callable, Optional, Type, Union

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.utils.text import slugify

from signoffs.core import models, utils
from signoffs.core.forms import SignoffFormsManager
from signoffs.core.renderers import SignoffRenderer
from signoffs.core.urls import SignoffUrlsManager

# type definitions shorts
opt_str = Union[bool, Optional[str]]
opt_callable = Union[Type, Callable]
signet_type = Union[str, Type[models.AbstractSignet]]
revoke_type = Union[str, Type[models.AbstractRevokedSignet]]


# The business logic for signing and revoking a signoff may be dependent on context that shouldn't be encoded in a Signoff
# Application logic may also need a way to sign or revoke without triggering Signoff permissions or signal logic.
# sign_signoff and revoke_signoff provide the default implementations for basic business logic for these operations.
# Implementors can override the behaviour by either overriding the class method or by injecting
#      a function with the same signature as the default implementations provided here...


def sign_signoff(signoff, user, commit=True, **kwargs):
    """
    Force signature onto given signoff for given user and save its signet, regardless of permissions or signoff state

    raises PermissionDenied otherwise (or whatever exception_type is given, e.g. ValidationError when saving forms)
    kwargs are passed directly to save - use commit=False to sign without saving.
    """
    signoff.signet.sign(user)
    signoff.signet.update(defaults=True, **signoff.get_signet_defaults(user))
    if commit:
        signoff.save(**kwargs)
    return signoff


def revoke_signoff(signoff, user, reason="", revokeModel=None, **kwargs):
    """
    Force revoke the given signoff for user regardless of permissions or signoff state.

    @param revokeModel: if supplied, create record of revocation, otherwise just delete the signet.
    """
    if revokeModel:
        return revokeModel.objects.create(
            signet=signoff.signet, user=user, reason=reason
        )
    else:
        signoff.signet.delete()
        signoff.signet.id = None


class DefaultSignoffBusinessLogic:
    """
    Defines the default business logic for signing and revoking a `Signoff` instance

    :::{note}
    `SignoffBusinessLogic` should *not* encode contextual (e.g., approval or approval process) business logic.
    :::
    """

    # Base permission and injectable logic for signing a signoff. Falsy for unrestricted
    perm: opt_str = ""  # e.g. 'signet.add_signet',
    sign_method: Callable = sign_signoff  # signing algorithm

    # Base permission and injectable logic for revoking a signoff. False to make irrevocable;  None (falsy) to use perm
    revoke_perm: opt_str = ""  # e.g. 'signet.delete_signet',
    revoke_method: Callable = revoke_signoff  # revoke signoffs algorithm

    def __init__(
        self, perm=None, sign_method=None, revoke_perm=None, revoke_method=None
    ):
        """Override default actions / permissions, or None to use class default"""
        self.perm = perm if perm is not None else self.perm
        self.sign_method = (
            sign_method or type(self).sign_method
        )  # don't bind sign_method to self here
        self.revoke_perm = revoke_perm if revoke_perm is not None else self.revoke_perm
        self.revoke_method = (
            revoke_method or type(self).revoke_method
        )  # don't bind revoke_method to self here

    # Make it easy to mix together business logic pieces in inline declarations
    @classmethod
    def mixin(cls, *mixins):
        """Return a subclass of this class with the given mixin classes mixed in"""

        # TODO: use functional syntax and build a suitable name for each class
        class MixedLogic(*mixins, cls):
            pass

        return MixedLogic

    # Signing Actions / Rules

    def is_permitted_signer(self, signoff_type, user):
        """Return True iff user has permission to sign a signoff of given type"""
        return (
            user is not None
            and user.id
            and (user.has_perm(self.perm) if self.perm else True)
        )

    def can_sign(self, signoff, user):
        """Return True iff the signoff instance can be signed by given user"""
        return not signoff.is_signed() and self.is_permitted_signer(type(signoff), user)

    def sign_if_permitted(self, signoff, user, commit=True, **kwargs):
        """
        Sign signoff for given user and save signet, if self.can_sign(user)
        raises PermissionDenied otherwise
        kwargs are passed directly to save - use commit=False to sign without saving.
        """
        if not self.can_sign(signoff, user):
            raise PermissionDenied(f"User {user} is not allowed to sign {signoff}")
        return self.sign(signoff, user, commit, **kwargs)

    def sign(self, signoff, user, commit=True, **kwargs):
        """
        Sign signoff for given user and save signet, regardless of permissions or signoff state - careful!

        kwargs are passed directly to save - use commit=False to sign without saving.
        """
        return self.sign_method(signoff, user, commit=commit, **kwargs)

    # Revoke Actions / Rules

    def is_permitted_revoker(self, signoff_type, user):
        """return True iff user has permission to revoke signoffs of given type"""
        revoke_perm = self.revoke_perm or self.perm
        return (
            False
            if self.revoke_perm is False
            else user.has_perm(revoke_perm)
            if revoke_perm
            else True
        )

    def can_revoke(self, signoff, user):
        """return True iff the signoff can be revoked by given user"""
        return signoff.is_signed() and self.is_permitted_revoker(type(signoff), user)

    def revoke_if_permitted(self, signoff, user, reason="", **kwargs):
        """Revoke the signoff for user if they have permission, otherwise raise PermissionDenied"""
        if not self.can_revoke(signoff, user):
            raise PermissionDenied(
                f"User {user} does not have permission to revoke {signoff.signet}"
            )
        return self.revoke(signoff, user, reason, **kwargs)

    def revoke(self, signoff, user, reason="", **kwargs):
        """Revoke the signoff regardless of permissions or signoff state - careful!"""
        kwargs.setdefault("revokeModel", signoff.revoke_model)
        return self.revoke_method(signoff, user, reason=reason, **kwargs)


class SignoffLogic(DefaultSignoffBusinessLogic):
    """Public API: Alias for `DefaultSignoffBusinessLogic"""


class AbstractSignoff:
    """
    Defines the abstract semantics for a Signoff and serves as base class for all concrete Signoff Types.

    A `Signoff Type` (i.e. subclass of `AbstractSignoff`) defines the behaviours for a specific type of signoff.

    A `Signoff` instance represents a timestamped agreement by a `User`, persisted by a `Signet` model instance.
    Signoffs are pure code objects, not stored in the DB - they define application logic, not application data.
    A Signoff Type defines:
      - how the `Signet` is labelled and rendered,
      - what permission is required to sign or revoke it,
      - what forms are used to sign and revoke them, etc.
    The default meta-data and services can be overridden in a subclass or passed to .register() factory.
    Signoff Types are registered in the `signoffs.registry.signoffs` registry where they can be retrieved by id
    :::{caution}
    `Signet` records are stored in DB with a reference to `Signoff.id`
     Be cautions not to change or delete id's that are in-use.
    :::
    """

    # id must be unique per type class, but human-legible / meaningful - dotted path recommended e.g. 'myapp.signoff'
    id: str = "signoff.abstract"  # unique identifier for type - used like FK, don't mess with these!

    # signetModel is required - every Signoff Type must supply a concrete Signet model to provide persistence layer
    signetModel: signet_type = None  # concrete model or 'app.model' string - REQUIRED

    # Signoff business logic, actions, and permissions
    logic: SignoffLogic = DefaultSignoffBusinessLogic()

    # revokeModel is optional - if provided, revoked signoff "receipts" will be kept, otherwise the signoff is deleted.
    revokeModel: revoke_type = None  # concrete model, if revoked signoffs are tracked

    # Define visual representation for signoffs of this Type. Label is a rendering detail, but common override.
    label: str = ""  # Label for (checkbox) form field, empty string for no label
    render: SignoffRenderer = SignoffRenderer()  # presentation logic service
    forms: SignoffFormsManager = SignoffFormsManager()  # Forms service
    urls: SignoffUrlsManager = SignoffUrlsManager()  # service to provide endpoints

    # Registration for Signoff Types (a.k.a. subclasses)

    @classmethod
    def register(cls, id, **kwargs):
        """
        Create, register, and return a new subclass of cls with overrides for given kwargs attributes

        Standard mechanism to define new Signoff types, typically in `my_app/models.py` or `my_app/signoffs.py`
        Usage:
        ```
            MySignoff = AbstractSignoff.register('my_signoff_type', label='Sign it!', ...)
        ```
        """
        from signoffs import registry

        class_name = utils.id_to_camel(id)
        kwargs["id"] = id
        signoff_type = type(class_name, (cls,), kwargs)
        registry.signoffs.register(signoff_type)
        return signoff_type

    @classmethod
    def validate(cls):
        """Run any class validation that must pass before class can be registered.  Invoked by registry."""
        if cls.signetModel is None:
            raise ImproperlyConfigured(
                f"Signoff Type {cls.id} must specify a Signet Model."
            )
        return True

    # Signoff Type accessors

    @classmethod
    def get_signetModel(cls):
        """Always use this accessor as the signetModel attribute may be an "app.Model" label"""
        if not cls.signetModel:
            raise ImproperlyConfigured(
                f"No Signet Model associated with Signoff {cls}."
            )
        if isinstance(cls.signetModel, str):
            cls.signetModel = apps.get_model(cls.signetModel)
        return cls.signetModel

    @classmethod
    def get_signet_queryset(cls):
        """Return a base (unfiltered) queryset of ALL signets for this Signoff Type"""
        return cls.get_signetModel().objects.filter(signoff_id=cls.id)

    @classmethod
    def get_revoked_signets_queryset(cls):
        """Return a base (unfiltered) queryset of ALL revoked signets for this Signoff Type"""
        return cls.get_signetModel().revoked_signets.filter(signoff_id=cls.id)

    @classmethod
    def get_revokeModel(cls):
        """Always use this accessor as the revokeModel attributed may be an "app.Model" label"""
        if isinstance(cls.revokeModel, str):
            cls.revokeModel = apps.get_model(cls.revokeModel)
        return cls.revokeModel

    @classmethod
    def get(cls, queryset=None, **filters):
        """
        Return the saved signoff that matches filters or a new signoff with these initial values if none exists

        Raises `MultipleObjectsReturned` if more than one signoff matches filter criteria.
        """
        SignetModel = cls.get_signetModel()
        queryset = queryset or SignetModel.objects.all()
        filters["signoff_id"] = cls.id
        try:
            return queryset.get(**filters).signoff
        except SignetModel.DoesNotExist:
            return cls(**filters)

    # Signoff Type behaviours

    @classmethod
    def create(cls, user, **kwargs):
        """Create and return a signoff signed by given user"""
        signoff = cls(**kwargs)
        signoff.sign(user=user)
        return signoff

    # Signoff instance behaviours

    def __init__(self, signet=None, subject=None, **kwargs):
        """
        Construct a Signoff instance backed by the given signet or an instance of `cls.signetModel(**kwargs)`

        `subject` is optional: the object this signoff is signing off on - set by `SignoffField` but otherwise unused.
        """
        if signet and kwargs:
            raise ImproperlyConfigured(
                "Construct signoff with either existing signet OR creation kwargs, not both."
            )
        self.signet = signet or self.get_new_signet(**kwargs)
        self._subject = subject
        if not self.signet.signoff_id == self.id:
            raise ImproperlyConfigured(
                f"Signoff Type {self} does not match Signet Model {self.signet.signoff_id}."
            )

    @property
    def subject(self):
        """
        The object being signed off on, if provided.

        Subclass with signet FK relations may want to override this to access the signet related object.
        `subject` is set by model Fields for convenient access to owner obj, but value is not used by core logic.
        """
        return self._subject

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def slug(self):
        """A slugified version of the signoff id, for places where a unique identifier slug is required"""
        return slugify(self.id)

    @property
    def signet_model(self):
        """Return the signoff model for this type"""
        return self.get_signetModel()

    @property
    def revoke_model(self):
        """Return the revoke model for this type"""
        return self.get_revokeModel()

    def get_new_signet(self, **initial_values):
        """Get a new, unsaved signet instance for this signoff type, with given optional initial values"""
        Signet = self.signet_model
        return Signet(
            signoff_id=self.id,
            **{
                k: v
                for k, v in initial_values.items()
                if k not in Signet.read_only_fields
            },
        )

    def __str__(self):
        return str(self.signet)

    def __eq__(self, other):
        return (
            isinstance(other, AbstractSignoff)
            and self.id == other.id
            and self.signet == other.signet
        )

    def matches(self, other):
        """Return True iff this signoff is of same type as other, which may be a signoff instance or a str"""
        return self.id == other if isinstance(other, str) else self.id == other.id

    def get_signet_defaults(self, user):
        """
        Return a dictionary of default values for fields this signoff's signet -

        Called during signoff.sign - provides default values that will NOT override values already set on the signet.
        See signets.get_signet_defaults for further docs.
        """
        return {}  # default implementation uses signets.get_signet_defaults

    # Signoff / Revoke Business Logic Delegation

    @classmethod
    def is_permitted_signer(cls, user):
        """return True iff user has permission to sign a signoff of this Type"""
        return cls.logic.is_permitted_signer(cls, user)

    def can_sign(self, user):
        """return True iff this signoff instance can be signed by given user"""
        return self.logic.can_sign(self, user)

    def sign_if_permitted(self, user, commit=True, **kwargs):
        """
        Sign for given user and save signet, if self.can_sign(user), raise `PermissionDenied` otherwise

        `kwargs` are passed directly to sign_method - use `commit=False` to sign without saving.
        """
        return self.logic.sign_if_permitted(self, user, commit=commit, **kwargs)

    def sign(self, user, commit=True, **kwargs):
        """
        Sign for given user and save signet, regardless of permissions or signoff state

        `kwargs` are passed directly to sign_method - use `commit=False` to sign without saving.
        """
        return self.logic.sign(self, user, commit=commit, **kwargs)

    @classmethod
    def is_permitted_revoker(cls, user):
        """Return True iff user has permission to revoke signoffs of this Type"""
        return cls.logic.is_permitted_revoker(cls, user)

    def can_revoke(self, user):
        """Return True iff this signoff can be revoked by given user"""
        return self.logic.can_revoke(self, user)

    def revoke_if_permitted(self, user, reason="", **kwargs):
        """Revoke this signoff for user if they have permission, otherwise raise PermissionDenied"""
        return self.logic.revoke_if_permitted(self, user, reason, **kwargs)

    def revoke(self, user, reason="", **kwargs):
        """Revoke this signoff regardless of permissions or signoff state - careful!"""
        return self.logic.revoke(self, user, reason, **kwargs)

    # Signet Delegation

    @property
    def signatory(self):
        """Return the user who signed, or AnonymousUser if signed but no signatory, None if not yet signed"""
        return self.signet.signatory

    @property
    def sigil(self):
        """Return the "sigil" on this signoff if it is signed, None otherwise"""
        return self.signet.sigil if self.is_signed() else None

    @property
    def sigil_label(self):
        """Return a label for the "sigil" on this signoff, if it is signed, None  otherwise"""
        return self.signet.sigil_label if self.is_signed() else None

    @property
    def timestamp(self):
        """Return the timestamp on this signoff if it is signed, None otherwise"""
        return self.signet.timestamp if self.is_signed() else None

    def has_user(self):
        """return True iff this signoff has a user-relation"""
        return self.signet.has_user()

    def can_save(self):
        """return True iff this signoff's signet is ready to be saved"""
        return self.signet.can_save() and self.can_sign(self.signet.user)

    def update(self, **attrs):
        """Update signet model fields with any attrs that match by name"""
        self.signet.update(**attrs)
        return self

    def validate_save(self):
        """Raise PermissionDenied if this Signoff cannot be saved, otherwise just pass."""
        if not self.can_sign(self.signet.user):
            raise PermissionDenied(
                f"User {self.signet.user} does not have permission to save {self}"
            )

    def save(self, *args, **kwargs):
        """Attempt to save a Signet with the provided associated data for this Signoff"""
        self.validate_save()
        self.signet.save(*args, **kwargs)
        return self

    def is_signed(self):
        """return True if this Signoff has a persistent representation"""
        return self.signet.is_signed()

    def is_revoked(self):
        """return True if this Signoff has been revoked"""
        return self.revokeModel and self.signet.is_revoked()

    @classmethod
    def has_object_relation(cls):
        return cls.get_signetModel().has_object_relation()


class BaseSignoff(AbstractSignoff):
    """
    A base Signoff Type to be used as abstract base class to register concrete Signoff Types
    Concrete Types will require a concrete Signet Model to back the signoff.
    :::{tip}
    Register new Signoff Types with:: `BaseSignoff.register(...)`
    :::
    """

    id = "signoffs.base-signoff"
    signetModel = None  # Concrete signetModel must be defined for registered Signoffs
    revokeModel = None  # revoking just deletes Signet unless a revokeModel is provided
    label = "I consent"


__all__ = [
    "sign_signoff",
    "revoke_signoff",
    "AbstractSignoff",
    "BaseSignoff",
    "DefaultSignoffBusinessLogic",
    "SignoffLogic",
]



"""
    Test Suite for regex_match pattern matching language
"""
from django.test import SimpleTestCase

from ..regex_match import (
    Pattern,
    PatternMatcher,
    all_of,
    exactly_n,
    exactly_one,
    in_series,
    n_or_more,
    one_of,
    one_or_more,
    wrap,
    zero_or_more,
    zero_or_one,
)


class RegexMatchTests(SimpleTestCase):
    def test_pattern_class(self):
        p = Pattern(regex="xyz", tokens=[1, 2, 3])
        self.assertEqual(p.regex, "xyz")
        self.assertEqual(p.tokens, [1, 2, 3])

    def test_wrap(self):
        self.assertEqual(wrap("A"), Pattern(regex="((?P<A>A) )", tokens=["A"]))
        self.assertEqual(
            wrap("A.B-C**"),
            Pattern(regex="((?P<ABC>A\\.B\\-C\\*\\*) )", tokens=["A.B-C**"]),
        )

    def test_exacly_one(self):
        self.assertEqual(
            exactly_one("A.B"), Pattern(regex="((?P<AB>A\\.B) )", tokens=["A.B"])
        )

    def test_zero_or_one(self):
        self.assertEqual(
            zero_or_one("A.B"), Pattern(regex="(((?P<AB>A\\.B) ){0,1})", tokens=["A.B"])
        )

    def test_zero_or_more(self):
        self.assertEqual(
            zero_or_more("A.B"), Pattern(regex="(((?P<AB>A\\.B) )*)", tokens=["A.B"])
        )

    def test_exactly_n(self):
        self.assertEqual(
            exactly_n("A.B", 2), Pattern(regex="((?P<AB>A\\.B) ){2,2}", tokens=["A.B"])
        )

    def test_one_or_more(self):
        self.assertEqual(
            one_or_more("A.B"), Pattern(regex="(((?P<AB>A\\.B) )+)", tokens=["A.B"])
        )
        self.assertEqual(
            one_or_more(Pattern(regex="xx", tokens=["xx"])),
            Pattern(regex="(xx+)", tokens=["xx"]),
        )

    def test_n_or_more(self):
        self.assertEqual(
            n_or_more("A.B", 2), Pattern(regex="(((?P<AB>A\\.B) ){2,})", tokens=["A.B"])
        )

    def test_in_series(self):
        self.assertEqual(
            in_series("A", "B"),
            Pattern(regex="(((?P<A>A) )((?P<B>B) ))", tokens=["A", "B"]),
        )
        self.assertEqual(
            in_series(exactly_one("A"), one_or_more("B")),
            Pattern(regex="(((?P<A>A) )(((?P<B>B) )+))", tokens=["A", "B"]),
        )

    def test_all_of(self):
        self.assertEqual(
            all_of("A", "B"),
            Pattern(
                regex="(((?P<A>A) )((?P<B>B) )|((?P<B>B) )((?P<A>A) ))",
                tokens=["A", "B"],
            ),
        )
        self.assertEqual(
            all_of("A", in_series("B", one_or_more("C"))),
            Pattern(
                regex="(((?P<A>A) )(((?P<B>B) )(((?P<C>C) )+))|(((?P<B>B) )(((?P<C>C) )+))((?P<A>A) ))",
                tokens=["A", "B", "C"],
            ),
        )

    def test_one_of(self):
        self.assertEqual(
            one_of("A", "B", "C"),
            Pattern(
                regex="(((?P<A>A) )|((?P<B>B) )|((?P<C>C) ))", tokens=["A", "B", "C"]
            ),
        )


class RegexPatternMatcherTests(SimpleTestCase):
    def assertMatch(self, m, is_valid, is_complete, next):
        self.assertEqual(m.is_valid, is_valid)
        self.assertEqual(m.is_complete, is_complete)
        self.assertSetEqual(set(m.next), set(next))

    def test_0(self):
        matcher = PatternMatcher(one_or_more("ABC"))
        m = matcher.match("")
        self.assertMatch(m, True, False, ["ABC"])
        m = matcher.match("ABC")
        self.assertMatch(m, True, True, ["ABC"])
        m = matcher.match("ABC ABC")
        self.assertMatch(m, True, True, ["ABC"])

    def test_1(self):
        matcher = PatternMatcher(one_or_more("A-B.C"))
        m = matcher.match("A-B.C")
        self.assertMatch(m, True, True, ["A-B.C"])
        m = matcher.match("A-B.C ")
        self.assertMatch(m, True, True, ["A-B.C"])
        m = matcher.match("B")
        self.assertMatch(m, False, False, [])

    def test_2(self):
        matcher = PatternMatcher(n_or_more("A*B.xy", 2))
        m = matcher.match("A*B.xy")
        self.assertMatch(m, True, False, ["A*B.xy"])
        m = matcher.match("A*B.xy A*B.xy")
        self.assertMatch(m, True, True, ["A*B.xy"])
        m = matcher.match("A*B.xy A*B.xy ")
        self.assertMatch(m, True, True, ["A*B.xy"])
        m = matcher.match("A*B.xy A*B.xy B*B.xy ")
        self.assertMatch(m, False, False, [])
        m = matcher.match("B A*B.xy A*B.xy")
        self.assertMatch(m, False, False, [])

    def test_3(self):
        matcher = PatternMatcher(exactly_one("A.B"))
        m = matcher.match("A.B ")
        self.assertMatch(m, True, True, [])
        m = matcher.match("A.B")
        self.assertMatch(m, True, True, [])
        m = matcher.match(".B")
        self.assertMatch(m, False, False, [])

    def test_4(self):
        matcher = PatternMatcher(exactly_n("A", 2))
        m = matcher.match("A")
        self.assertMatch(m, True, False, ["A"])
        m = matcher.match("A A")
        self.assertMatch(m, True, True, [])
        m = matcher.match("B")
        self.assertMatch(m, False, False, [])

    def test_5(self):
        matcher = PatternMatcher(
            in_series(one_or_more("A.B"), exactly_one("B-B"), n_or_more("C*C", 2))
        )
        m = matcher.match("A.B")
        self.assertMatch(m, True, False, ["A.B", "B-B"])
        m = matcher.match("A.B B-B")
        self.assertMatch(m, True, False, ["C*C"])
        m = matcher.match("A.B A.B B-B")
        self.assertMatch(m, True, False, ["C*C"])
        m = matcher.match("A.B A.B B-B C*C C*C")
        self.assertMatch(m, True, True, ["C*C"])
        m = matcher.match("B")
        self.assertMatch(m, False, False, [])

    def test_6(self):
        matcher = PatternMatcher(
            all_of(one_or_more("A"), exactly_one("B"), n_or_more("C", 2))
        )
        m = matcher.match("A")
        self.assertMatch(m, True, False, ["A", "B", "C"])
        m = matcher.match("B")
        self.assertMatch(m, True, False, ["A", "C"])
        m = matcher.match("C")
        self.assertMatch(m, True, False, ["C"])
        m = matcher.match("C B A")
        self.assertMatch(m, False, False, [])
        m = matcher.match("B C C A A")
        self.assertMatch(m, True, True, ["A"])
        m = matcher.match("C C A B")
        self.assertMatch(m, True, True, [])
        m = matcher.match("B C A B")
        self.assertMatch(m, False, False, [])

    def test_7(self):
        matcher = PatternMatcher(
            all_of(one_or_more("A"), in_series(exactly_one("B"), n_or_more("C", 2)))
        )
        m = matcher.match("A")
        self.assertMatch(m, True, False, ["A", "B"])
        m = matcher.match("B C")
        self.assertMatch(m, True, False, ["C"])
        m = matcher.match("C")
        self.assertMatch(m, False, False, [])

    def test_8(self):
        matcher = PatternMatcher(
            in_series(one_or_more("A"), all_of(exactly_one("B"), n_or_more("C", 2)))
        )
        m = matcher.match("A")
        self.assertMatch(m, True, False, ["A", "B", "C"])
        m = matcher.match("A A C")
        self.assertMatch(m, True, False, ["C"])
        m = matcher.match("A B C")
        self.assertMatch(m, True, False, ["C"])
        m = matcher.match("A C C B")
        self.assertMatch(m, True, True, [])

    def test_9(self):
        matcher = PatternMatcher(
            all_of(exactly_one("A"), n_or_more("B", 2), one_or_more("C"))
        )
        m = matcher.match("C B B B A")
        self.assertMatch(m, True, True, [])

    def test_10(self):
        matcher = PatternMatcher(
            one_of(exactly_one("A"), n_or_more("B", 2), one_or_more("C"))
        )
        m = matcher.match("A")
        self.assertMatch(m, True, True, [])
        m = matcher.match("B B")
        self.assertMatch(m, True, True, ["B"])
        m = matcher.match("C")
        self.assertMatch(m, True, True, ["C"])
        m = matcher.match("A B")
        self.assertMatch(m, False, False, [])

    def test_11(self):
        matcher = PatternMatcher(zero_or_one("A"))
        m = matcher.match("")
        self.assertMatch(
            m, True, False, ["A"]
        )  # No pattern is complete without at least one token!
        m = matcher.match("A")
        self.assertMatch(m, True, True, [])
        m = matcher.match("A A")
        self.assertMatch(m, False, False, [])

    def test_12(self):
        matcher = PatternMatcher(zero_or_more("A"))
        m = matcher.match("")
        self.assertMatch(m, True, False, ["A"])
        m = matcher.match("A")
        self.assertMatch(m, True, True, ["A"])
        m = matcher.match("A A")
        self.assertMatch(m, True, True, ["A"])



"""
signoffs.utils test suite
"""
from types import SimpleNamespace

from django.test import SimpleTestCase

from signoffs.core import utils


class UtilsTests(SimpleTestCase):
    def test_regexs(self):
        self.assertEqual(
            utils.split_caps_run.sub(r"\1_\2", "abstractHTTPResponseCode"),
            "abstractHTTP_ResponseCode",
        )

    def test_camel_to_snake(self):
        self.assertEqual(
            utils.camel_to_snake("camel2_camel2_case"), "camel2_camel2_case"
        )
        self.assertEqual(utils.camel_to_snake("NormalClassName"), "normal_class_name")
        self.assertEqual(
            utils.camel_to_snake("abstractHTTPResponseCode"),
            "abstract_http_response_code",
        )
        self.assertEqual(
            utils.camel_to_snake("HTTPResponseCodeXYZ"), "http_response_code_xyz"
        )

    def test_id_to_camel(self):
        self.assertEqual(utils.id_to_camel("NormalClassName"), "NormalClassName")
        self.assertEqual(utils.id_to_camel("snake_snake_case"), "SnakeSnakeCase")
        self.assertEqual(utils.id_to_camel("dot.separated-id"), "DotSeparatedId")


data = SimpleNamespace(
    obj1=SimpleNamespace(
        attr1="Dent",
        attr2=42,
    ),
    obj2=SimpleNamespace(obj=SimpleNamespace(attr1="Route", attr2=66)),
)


class AccessorTests(SimpleTestCase):
    def test_basic_resolve(self):
        x = utils.Accessor("obj1__attr1")
        self.assertEqual(x.resolve(data), "Dent")
        x = utils.Accessor("obj1__attr2")
        self.assertEqual(x.resolve(data), 42)

    def test_penultimate(self):
        x = utils.Accessor("obj2__obj__attr2")
        self.assertEqual(x.resolve(data), 66)
        w, r = x.penultimate_accessor()
        self.assertEqual((w, r), ("obj2__obj", "attr2"))
        self.assertEqual(w.resolve(data), data.obj2.obj)
        v, r = w.penultimate_accessor()
        self.assertEqual((v, r), ("obj2", "obj"))
        self.assertEqual(v.resolve(data), data.obj2)
        u, r = v.penultimate_accessor()
        self.assertEqual((u, r), ("", "obj2"))
        self.assertEqual(u.resolve(data), None)



"""
A form for collecting approval signoffs
"""
from django import forms

from signoffs.core.forms import AbstractSignoffForm

from .models import Signet as ApprovalSignet


class ApprovalSignoffForm(AbstractSignoffForm):
    """Form for collecting approval signoffs"""

    class Meta(AbstractSignoffForm.Meta):
        model = ApprovalSignet
        widgets = {
            "stamp": forms.HiddenInput,
        }



"""
    Proxy for Signoff Types to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""
from django.apps import apps

from .core import signing_order

from .core import utils

from .core.forms import (
    SignoffFormsManager,
    SignoffTypeForms,
)
from .core.renderers import (
    SignoffInstanceRenderer,
    SignoffRenderer,
)
from .core.signoffs import (
    AbstractSignoff,
    BaseSignoff,
    SignoffLogic,
)
from .core.urls import (
    SignoffInstanceUrls,
    SignoffUrlsManager,
)

if apps.is_installed("signoffs.contrib.signets"):
    from .contrib.signets.signoffs import (
        IrrevokableSignoff,
        RevokableSignoff,
        SimpleSignoff,
    )



from django.conf import settings
from django.urls import include, path

import signoffs.urls
from tests.test_app import views

# ensure signoff_id converter is registered - how should a reusable app ensure its converters are registered?
SignoffIdConverter = signoffs.urls.SignoffIdConverter


# Views for interactive testing
urlpatterns = [
    path(
        "signoff/<id:signoff_id>/<int:pk>/detail/",
        view=views.SignoffDetailView.as_view(),
        name="detail",
    ),
    path(
        "approval/<id:approval_id>/<int:pk>/detail/",
        view=views.ApprovalDetailView.as_view(),
        name="detail",
    ),
    path("", include("signoffs.urls")),
    path("auth/", include("django.contrib.auth.urls")),
]


if settings.DEBUG:
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns

    urlpatterns += staticfiles_urlpatterns()



"""
App-dependent tests for signoff models
"""
from django.test import TestCase

from signoffs.core.tests import fixtures
from signoffs.signoffs import SignoffLogic
from tests.test_app import models, signoffs


class SignoffTests(TestCase):
    def test_default_sigil(self):
        u = fixtures.get_user()
        o = models.Signet(signoff_id="test_app.agree", user=u)
        o.save()
        self.assertEqual(o.sigil, u.get_full_name())

    def test_custom_sigil(self):
        u = fixtures.get_user()
        o = models.Signet(signoff_id="test_app.agree", user=u, sigil="Bugs Bunny")
        o.save()
        self.assertEqual(o.sigil, "Bugs Bunny")


class SignoffTypePermissionTests(TestCase):
    def test_can_sign_no_perm(self):
        s = signoffs.agree_signoff
        u = fixtures.get_user()
        self.assertTrue(s.is_permitted_signer(u))

    def test_can_sign_perm(self):
        perm = fixtures.get_perm("can_sign")
        s = signoffs.agree_signoff
        u = fixtures.get_user(perms=(perm,))
        self.assertTrue(s.is_permitted_signer(u))

    def test_cant_sign_without_permission(self):
        perm = fixtures.get_perm("no_signie")
        s = signoffs.TestSignoff.register(
            id="test3.signoff", logic=SignoffLogic(perm="auth.can_sign")
        )
        u = fixtures.get_user(perms=(perm,))
        self.assertFalse(s.is_permitted_signer(u))

    def test_can_revoke_no_perm(self):
        s = signoffs.TestSignoff.register(
            id="test4.signoff", logic=SignoffLogic(perm=None)
        )
        u = fixtures.get_user()
        self.assertTrue(s.is_permitted_revoker(u))

    def test_can_revoke_with_perm(self):
        perm = fixtures.get_perm("can_sign")
        revoke = fixtures.get_perm("can_revoke")
        s = signoffs.TestSignoff.register(
            id="test5.signoff", logic=SignoffLogic(perm="auth.can_sign")
        )
        u = fixtures.get_user(perms=(perm, revoke))
        self.assertTrue(s.is_permitted_revoker(u))
        s2 = signoffs.TestSignoff.register(
            id="test5.signoff2",
            logic=SignoffLogic(perm="auth.can_sign", revoke_perm="auth.can_revoke"),
        )
        self.assertTrue(s2.is_permitted_revoker(u))

    def test_cant_revoke_without_permission(self):
        not_perm = fixtures.get_perm("no_signie")
        s = signoffs.TestSignoff.register(
            id="test6.signoff", logic=SignoffLogic(perm="auth.can_sign")
        )
        u = fixtures.get_user(perms=(not_perm,))
        self.assertFalse(s.is_permitted_revoker(u))
        perm = fixtures.get_perm("can_sign")
        u2 = fixtures.get_user(
            perms=(
                not_perm,
                perm,
            )
        )
        self.assertTrue(s.is_permitted_revoker(u2))
        s2 = signoffs.TestSignoff.register(
            id="test6.signoff2",
            logic=SignoffLogic(perm="auth.can_sign", revoke_perm="auth.can_revoke"),
        )
        self.assertFalse(s2.is_permitted_revoker(u2))


class SignoffPermissionTests(TestCase):
    def test_can_sign_no_perm(self):
        u = fixtures.get_user()
        o = models.Signet(signoff_id="test_app.agree", user=u)
        self.assertTrue(o.can_save())

    def test_can_sign_perm(self):
        perm = fixtures.get_perm("can_sign")
        u = fixtures.get_user(perms=(perm,))
        o = models.Signet(signoff_id="test_app.consent", user=u)
        self.assertTrue(o.can_save())

    def test_cant_sign_without_permission(self):
        perm = fixtures.get_perm("no_signie")
        u = fixtures.get_user(perms=(perm,))
        signoff = signoffs.consent_signoff(user=u)
        self.assertFalse(signoff.can_save())

    def test_can_revoke_no_perm(self):
        u = fixtures.get_user()
        signoff = signoffs.agree_signoff(user=u)
        self.assertTrue(signoff.can_save())

    def test_can_revoke_with_perm(self):
        u = fixtures.get_user(perms=("can_sign", "can_revoke", "can_accept"))
        signoff = signoffs.consent_signoff(user=u).save()
        self.assertTrue(signoff.can_revoke(u))
        signoff2 = signoffs.agree_signoff(user=u).save()
        self.assertTrue(signoff2.can_revoke(u))

    def test_cant_revoke_without_permission(self):
        u = fixtures.get_user(perms=("no_signie",))
        signoff = signoffs.consent_signoff(user=u)
        self.assertFalse(signoff.can_revoke(u))
        perm = fixtures.get_perm("can_sign")
        u2 = fixtures.get_user(
            perms=(
                "no_signie",
                perm,
            )
        )
        signoff2 = signoffs.consent_signoff(user=u2)
        self.assertTrue(signoff2.can_save())
        self.assertFalse(signoff2.can_revoke(u2))


class SignoffRevokeTests(TestCase):
    def test_revoke_no_model(self):
        u = fixtures.get_user()
        signoff = signoffs.agree_signoff(user=u)
        signoff.save()
        signoff.revoke(u, reason="why not")
        self.assertEqual(models.Signet._base_manager.count(), 0)

    def test_revoke_with_model(self):
        perm = fixtures.get_perm("can_review")
        u = fixtures.get_user(perms=(perm,))
        r = models.Report.objects.create(contents="Awesomeness in textual form.")
        signoff = r.signoffs.create(user=u)
        signoff.revoke(u, reason="why not")
        self.assertEqual(models.ReportSignet._base_manager.count(), 1)
        self.assertEqual(models.ReportSignet.objects.count(), 0)
        self.assertEqual(models.RevokeReportSignet.objects.count(), 1)
        revoked = models.RevokeReportSignet.objects.select_related("signet").get(
            signet=signoff.signet
        )
        self.assertEqual(revoked.signet, signoff.signet)


class SignoffRelationsTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        perm = fixtures.get_perm("can_review")
        u = fixtures.get_user(perms=(perm,))
        r = models.Report.objects.create(contents="Awesomeness in textual form.")
        so = r.signoffs.create(user=u)
        so2 = r.signoffs.create(user=u)
        cls.signets = [so.signet, so2.signet]
        cls.u = u
        cls.r = r

    def test_signet_set(self):
        self.assertEqual(self.r.signatories.all().count(), len(self.signets))
        self.assertQuerysetEqual(self.r.signatories.order_by("pk"), self.signets)

    def test_signoff_set(self):
        self.assertEqual(self.r.signoffs.count(), len(self.signets))
        self.assertSetEqual(
            {so.signet for so in self.r.signoffs.all()}, set(self.signets)
        )

    def test_signet_set_revoke(self):
        signoff = self.signets[0].signoff
        signoff.revoke(self.u, reason="why not")
        self.assertEqual(self.r.signatories.all().count(), len(self.signets) - 1)
        self.assertEqual(self.r.signatories.all().first(), self.signets[1])
        # Check the revoked instance
        self.assertEqual(signoff.signet.revoked.user, self.u)
        self.assertEqual(signoff.signet.revoked.signet, signoff.signet)
        # but the signet is still there, just not visible b/c it's been revoked.
        self.assertEqual(self.r.signatories(manager="revoked_signets").count(), 1)
        self.assertEqual(
            models.ReportSignet._base_manager.filter(report=self.r).count(),
            len(self.signets),
        )

    def test_revoked_signoff_set(self):
        n_revoked = 3
        for _ in range(n_revoked):
            so = self.r.signoffs.create(self.u)
            so.revoke(self.u, reason="just because")
        self.assertEqual(self.r.signoffs.count(), len(self.signets))
        revoked = self.r.signoffs.revoked()
        self.assertEqual(len(revoked), n_revoked)
        self.assertTrue(all(signet.is_signed() for signet in revoked))
        self.assertTrue(all(signet.is_revoked() for signet in revoked))
        self.assertTrue(all(signet.revoked.user == self.u for signet in revoked))



"""
Proxy for signoffs.core.forms to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""
from django.apps import apps

from signoffs.core.forms import (
    AbstractSignoffForm,
    AbstractSignoffRevokeForm,
    SignoffFormsManager,
    SignoffTypeForms,
    revoke_form_factory,
    signoff_form_factory,
)

if apps.is_installed("signoffs.contrib.approvals"):
    from signoffs.contrib.approvals.forms import ApprovalSignoffForm



"""
    Simplify rendering signoff objects in templates
"""
from django import template

register = template.Library()


@register.filter
def can_revoke(signoff_or_approval_instance, user):
    """Returns True if the signoff_or_approval_instance can be revoked by the given user (in theory)"""
    return signoff_or_approval_instance.can_revoke(user)


@register.filter
def next_signoffs_for_user(approval_instance, user):
    """Returns list of next signoffs available for signing by user"""
    return approval_instance.next_signoffs(for_user=user)


@register.filter
def next_signoff_for_user(approval_instance, user):
    """Returns the next signoffs avaialble for signing by user"""
    return approval_instance.get_next_signoff(for_user=user)


@register.simple_tag(takes_context=True)
def render_signoff(context, signoff_instance, action="__call__", **kwargs):
    """
    Render the signoff_instance using the given method (action) of its renderer.

    kwargs are passed to action method
    """
    method = getattr(signoff_instance.render, action, signoff_instance.render.__call__)
    return method(**kwargs, context=context)


def _get_request_user(context, **kwargs):
    """return user object from context or kwargs, or None"""
    # Only need the request.user, so don't require a request object, but often convenient to use one  ** sigh **
    request_user = kwargs.get(
        "request_user", context.get("request_user", context.get("user", None))
    )
    request = kwargs.get("request", context.get("request", None))
    return request_user or (request.user if request else None)


@register.simple_tag(takes_context=True)
def render_approval_signoff(
    context, approval, signoff_instance, action="__call__", **kwargs
):
    """
    Render the signoff_instance within context of approval using the given method (action) of its renderer.

    kwargs are passed to action method
    """
    user = _get_request_user(context, **kwargs)
    if not approval.can_revoke_signoff(signoff_instance, user):
        kwargs["show_revoke"] = False
    return render_signoff(context, signoff_instance, action=action, **kwargs)


@register.simple_tag(takes_context=True)
def render_approval(context, approval_instance, **kwargs):
    """
    Render the approval_instance

    kwargs are passed to renderer.render method
    """
    return approval_instance.render(**kwargs, context=context)


@register.simple_tag(takes_context=True)
def render_process_approval(context, approval_process, approval_instance, **kwargs):
    """
    Render the approval_instance within context of approval_process.

    kwargs are passed to render method
    """
    if not approval_process.is_revokable_approval(approval_instance):
        kwargs["show_revoke"] = False
    return approval_instance.render(**kwargs, context=context)


@register.simple_tag(takes_context=True)
def render_approval_process(context, approval_process, **kwargs):
    """
    Render the approval_process

    kwargs are passed to render method
    """
    return approval_process.render(**kwargs, context=context)



"""
It can be very convenient for an approval to know the endpoint used to DELETE its `Stamp`.
In fact, the default renderers assume an approval knows the URL for revoke requests.

This component can be extended to provide flexible url services to approval instances.
They are generally "injected" into the Approval Type using a `ApprovalUrlsManager` service descriptor
"""
from django.urls import reverse

from signoffs.core.utils import service


class ApprovalInstanceUrls:
    """
    Defines the urls for Revoking an Approval instance
    """

    # Define URL patterns for revoking approvals
    revoke_url_name: str = ""

    def __init__(self, approval_instance, revoke_url_name=None):
        """Override default actions, or leave parameter None to use class default"""
        self.approval = approval_instance
        self.revoke_url_name = revoke_url_name or self.revoke_url_name

    def get_revoke_url(self, args=None, kwargs=None):
        """Return the URL for requests to revoke the approval"""
        args = args or [
            self.approval.stamp.pk,
        ]
        kwargs = kwargs or {}
        return (
            reverse(self.revoke_url_name, args=args, kwargs=kwargs)
            if self.revoke_url_name
            else ""
        )


class ApprovalUrlsManager(service(ApprovalInstanceUrls)):
    """
    A descriptor class that "injects" a ApprovalInstanceUrls instance into a Approval instance.

    To inject custom URL services:
      - provide a custom service_class:  `urls=ApprovalUrlsManager(service_class=MyInstanceUrls)`
      - OR specialize class attributes:
        `MyApprovalUrlsManager = utils.service(ApprovalInstanceUrls, revoke_url_name="my:app:revoke_view")`
      - OR both... `ApprovalUrlsManager = utils.service(MyInstanceUrls)`

    """



"""
    Register some signoffs for testing with
    Will be auto-discovered without need to import this module.
"""
from signoffs.signoffs import BaseSignoff, RevokableSignoff, SignoffLogic, SimpleSignoff


class TestSignoff(BaseSignoff):
    signetModel = "signoffs_signets.Signet"
    label = "This is agreeable"


agree_signoff = SimpleSignoff.register(id="test_app.agree")
consent_signoff = RevokableSignoff.register(
    id="test_app.consent",
    logic=SignoffLogic(perm="auth.can_sign", revoke_perm="auth.can_revoke"),
)

accept_signoff = SimpleSignoff.register(
    id="test_app.accept",
    signetModel="test_app.ReportSignet",
    label="I Accept",
    logic=SignoffLogic(perm="auth.can_accept"),
)

report_signoff = RevokableSignoff.register(
    id="test_app.report_signoff",
    signetModel="test_app.ReportSignet",
    revokeModel="test_app.RevokeReportSignet",
    label="Reviewed",
    logic=SignoffLogic(perm="auth.can_review"),
)

hr_signoff = BaseSignoff.register(
    id="test_app.hr_signoff",
    signetModel="test_app.VacationSignet",
    label="Vacation Approved",
    logic=SignoffLogic(perm="auth.can_approve_hr"),
)

mngr_signoff = BaseSignoff.register(
    id="test_app.mngr_signoff",
    signetModel="test_app.VacationSignet",
    label="Vacation Approved",
    logic=SignoffLogic(perm="auth.can_approve_mngr"),
)



"""
A `RequestActions` object is a facade that presents a simplified interface for handling requests to
sign and revoke signoff objects.

There are just a couple "actions" that can be taken with a signoff
  - it gets signed, it gets revoked, that's about it.
But these simple actions often initiate a cascade of notifications and/or state transitions.
The cascade of actions taken during a signoff request will be application dependent to some degree, so the
`RequestActions` classes are designed to be open for extension.

A `RequestActionsProtocol` is an API that a generic signoff- or approval-processing view
can use to handle a request.
Concrete `Actions` encapsulate the business logic for requests made for a particular signoff or approval process.
Thus, they provide reusable, extensible, composable View logic, outside the Views hierarchy.

:::{tip}
For best cohesion: the `Signoff.logic` and `Approval.logic` objects implement all business rules
while an `Action` object just orchestrates them - resist temptation to encode business rules in Actions!
:::

The protocols, default implementations, and actions provided here are sufficient to handle a range of basic use-cases.
They are not intended to be exhaustive - consider them as a skeleton to demonstrate how a signoff request might
typically be handled.
"""
from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import Callable, Protocol

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction

import signoffs.registry
from signoffs.approvals import AbstractApproval
from signoffs.process import ApprovalsProcess
from signoffs.signoffs import AbstractSignoff

User = get_user_model()

##################
#  SIGNOFF ACTIONS
##################


class SignoffValidator(Protocol):
    """Basic API to validate a signoff request for a specific user"""

    user: User

    def is_valid_signoff_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be signed by user."""
        ...

    def is_valid_revoke_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be revoked by user."""
        ...


@dataclass
class BasicSignoffValidator:
    """
    Default Signoff request permissions validation logic

    Implements`SignoffValidator` Protocol
    """

    user: User
    verify_signet: Callable[[AbstractSignoff], bool] = lambda signoff: True
    """Optional function to validate request data against a signoff, or None to skip"""

    def is_valid_signoff_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be signed by user."""
        return signoff.can_sign(self.user)

    def is_valid_revoke_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be revoked by user."""
        return signoff.can_revoke(self.user) and self.verify_signet(signoff)


class SignoffCommitter(Protocol):
    """Basic API to commit a signoff request to DB for a specific user"""

    user: User

    def sign(self, signoff: AbstractSignoff):
        """Sign and commit the signoff for given user - no validation, just do it!"""
        ...

    def revoke(self, signoff: AbstractSignoff):
        """Revoke the signoff for given user and commit changes to DB - no validation, just do it!"""
        ...


@dataclass
class BasicSignoffCommitter:
    """
    Default Signoff request commit logic

    Implements `SignoffCommitter` Protocol
    """

    user: User

    def sign(self, signoff: AbstractSignoff):
        """Sign and commit the signoff for given user - no validation, just do it!"""
        signoff.sign(self.user)

    def revoke(self, signoff: AbstractSignoff):
        """Revoke the signoff for given user and commit changes to DB - no validation, just do it!"""
        signoff.revoke(self.user)


class SignoffRequestFormHandler(Protocol):
    """Basic API to provision and validate forms from signoff request data"""

    data: dict

    def get_signoff_form(self):
        """Return the signoff form bound to data or None if no valid signoff form could be constructed"""
        ...

    def get_signed_signoff(self, user: User) -> AbstractSignoff:
        """Validate data against signoff form, return signed but unsaved signoff or None if form doesn't validate"""
        ...

    def get_revoke_form(self):
        """Return a revoke form bound to data or None if no valid signoff form could be constructed"""
        ...

    def get_revoked_signoff(self, user: User) -> AbstractSignoff:
        """Validate data against revoke form, return unrevoked signoff or None if form doesn't validate"""
        ...


@dataclass
class BasicSignoffFormHandler:
    """
    Provision and validate a signoff form from a data source

    Implements `SignoffRequestFormHandler` Protocol
    """

    data: dict
    """GET or POST data dictionary"""
    signoff_subject: object = None
    """optional subject of the signoff - signoff.subject set if provided"""
    signoff_id_key: str = "signoff_id"
    """default name of data key (i.e., form field) for retrieving signoff_id from data dict"""

    def get_signoff_type(self):
        """Return the signoff type indicated by `data[signoff_id_key]`, or None"""
        signoff_id = self.data.get(self.signoff_id_key, None)
        try:
            return signoffs.registry.get_signoff_type(signoff_id)
        except ImproperlyConfigured:  # invalid signoff_id in the data
            return None

    def get_signoff_form(self):
        """Return the signoff form bound to data or None if no valid signoff form could be constructed"""
        signoff_type = self.get_signoff_type()
        return signoff_type.forms.get_signoff_form(self.data) if signoff_type else None

    def get_signed_signoff(self, user):
        """Validate data against signoff form, return signed but unsaved signoff or None if form doesn't validate"""
        signoff_form = self.get_signoff_form()
        if not signoff_form or not signoff_form.is_valid():
            return None
        signoff = signoff_form.sign(user=user, commit=False)
        if signoff and self.signoff_subject:
            signoff.subject = self.signoff_subject
        return signoff

    def get_revoke_form(self):
        """Return a revoke form bound to data or None if no valid signoff form could be constructed"""
        signoff_type = self.get_signoff_type()
        return signoff_type.forms.get_revoke_form(self.data) if signoff_type else None

    def get_revoked_signoff(self, user):
        """Validate data against revoke form, return unrevoked signoff or None if form doesn't validate"""
        revoke_form = self.get_revoke_form()
        if not revoke_form or not revoke_form.is_valid():
            return None
        return revoke_form.revoke(user=user, commit=False)


class SignoffRequestActions(Protocol):
    """Basic API to process a signoff request"""

    signoff: AbstractSignoff

    def sign_signoff(self, commit=True):
        """Handle request to sign the signoff, return True iff signoff was signed"""
        ...

    def revoke_signoff(self, commit=True):
        """Handle request to revoke the signoff, return True iff signoff was revoked"""
        ...


def verify_consistent_signet_id(signoff, request_signet_id=None) -> bool:
    """As an extra data integrity check, revoke URL's may also contain the signet_id - verify it matches"""
    return (request_signet_id == signoff.signet.pk) if request_signet_id else True


def get_verify_signet(kwargs, signet_id_key="signet_id"):
    """Helper to return a `verify_signet` function with correct signature for use with `BasicSignoffValidator`"""
    return partial(
        verify_consistent_signet_id, request_signet_id=kwargs.get(signet_id_key, None)
    )


class BasicUserSignoffActions:
    """
    Concrete SignoffRequestActions to handle simple signoff requests that don't trigger any follow-up actions

    An extensible base class for extending this simple case that
    implements `SignoffRequestActions` Protocol
    Facade: delegation to signoff's `forms` and business `logic`
    """

    form_handler_class: SignoffRequestFormHandler = BasicSignoffFormHandler
    validator_class: SignoffValidator = BasicSignoffValidator
    committer_class: SignoffCommitter = BasicSignoffCommitter

    def __init__(
        self,
        user: User,
        data: dict,
        form_handler: SignoffRequestFormHandler = None,
        validator: SignoffValidator = None,
        committer: SignoffCommitter = None,
        **kwargs,
    ):
        """
        Define actions available to the given user based on request data

        :param user: typically the request.user
        :param dict data: a dict-like object, typically with the GET or POST data from the request
        :param kwargs: extra kwargs, typically view.kwargs, for validating and provisioning objects
        """
        self.user = user
        self.data = data
        self.forms = form_handler or self.form_handler_class(data)
        self.validator = validator or self.validator_class(
            user=user, verify_signet=get_verify_signet(kwargs)
        )
        self.committer = committer or self.committer_class(user=user)
        self.kwargs = kwargs
        self.signoff = None  # populated by calling sign_ or revoke_signoff

    def verify_consistent_signet_id(self, signoff) -> bool:
        """As an extra data integrity check, revoke URL's may also contain the signet_id - check it matches"""
        signet_id = self.kwargs.get("signet_id", None)
        return (signet_id == signoff.signet.pk) if signet_id else True

    # "Template Method" hooks: to extend / override sign_signoff without duplicating core algorithm

    def signoff_success(self, signoff):
        """Return value for successful signoff."""
        return True

    def signoff_failed(self, signoff):
        """Return value for failed signoff."""
        return False

    # Core algorithm:  Provision a signoff form, validate, and attempt to sign it.

    def sign_signoff(self, commit=True):
        """
        Handle request to sign a signoff form defined in `data`, return True iff signoff was signed

        :param bool commit: False to validate the form and sign self.signoff, but not commit signet to DB.
        """
        self.signoff = self.forms.get_signed_signoff(self.user)
        if self.signoff and self.validator.is_valid_signoff_request(self.signoff):
            if commit:
                self.committer.sign(self.signoff)
            return self.signoff_success(self.signoff)
        return self.signoff_failed(self.signoff)

    # "Template Method" hooks: to extend / override revoke_signoff without duplicating core algorithm

    def revoke_success(self, signoff):
        """Return value for successful revoke."""
        return True

    def revoke_failed(self, signoff):
        """Return value for failed revoke."""
        return False

    # Core algorithm:  Provision a revoke form, validate, and attempt to revoke the signet.

    def revoke_signoff(self, commit=True):
        """
        Handle request to revoke given signoff, return True iff signoff was revoked

        :param bool commit: False to validate the form, but not actually revoke the signet and commit to DB.

        :::{note}
        `commit=False` means "perform validation and return what would have happened, but do not actually revoke"
        There is no way to "revoke" a signoff without committing the change.  Call again to do the actual revoke.
        :::
        """
        self.signoff = self.forms.get_revoked_signoff(self.user)

        if self.signoff and self.validator.is_valid_revoke_request(self.signoff):
            if commit:
                self.committer.revoke(self.signoff)
            return self.revoke_success(self.signoff)
        return self.revoke_failed(self.signoff)


##################
# APPROVAL ACTIONS
##################


class ApprovalSignoffValidator(BasicSignoffValidator):
    """
    Default permissions validation logic for requests to Signoff on an Approval

    Implements`SignoffValidator` Protocol
    """

    def __init__(
        self,
        user: User,
        approval: AbstractApproval,
        verify_signet: Callable[[AbstractSignoff], bool] = lambda signoff: True,
        verify_stamp: Callable[[AbstractSignoff], bool] = lambda signoff: True,
    ):
        """
        Initialize validator for signoffs on the given approval.

        :param AbstractApproval approval: that approval for which to validate signoff requests
        :param Callable verify_signet: Optional function to validate request data against a signoff, or None to skip
        :param Callable verify_stamp: Optional function to validate request data, or None to skip
        """
        super().__init__(user=user, verify_signet=verify_signet)
        self.approval = approval
        self.verify_stamp = verify_stamp

    def is_valid_signoff_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be signed by user."""
        return (
            super().is_valid_signoff_request(signoff)
            and self.verify_stamp(signoff)
            and self.approval.can_sign(self.user, signoff)
        )

    def is_valid_revoke_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be revoked by user."""
        return (
            super().is_valid_revoke_request(signoff)
            and self.verify_stamp(signoff)
            and self.approval.can_revoke_signoff(signoff, self.user)
        )


@dataclass
class ApprovalSignoffCommitter(BasicSignoffCommitter):
    """
    Default commit logic for requests to Signoff on an Approval.

    Approves and commits the approval if it is ready after the current signoff is committed.
    :::{caution}
    Potential sync. issue here if approval has cached / prefetched signoffs - don't do that!
    :::
    """

    approval: AbstractApproval
    approve_if_ready: Callable[[AbstractApproval], bool]

    def sign(self, signoff: AbstractSignoff):
        """Sign and commit the signoff, and approve approval if ready - no validation, just do it!"""
        with transaction.atomic():
            super().sign(signoff)
            self.approve_if_ready(self.approval)


class ApprovalRequestActions(Protocol):
    """Basic API to process an approval request"""

    signoff: AbstractSignoff
    approval: AbstractApproval

    def approve(self, commit=True):
        """Handle request to approve the approval, return True iff it was approved."""
        ...

    def revoke_approval(self, commit=True):
        """Handle request to revoke approval, return True iff approval was revoked"""
        ...


def verify_consistent_stamp_id(approval, signoff=None, request_stamp_id=None):
    """
    Return True iff data relations for approval, signoff, and request are self-consistent

    Validates that signoff.stamp, if it exists, is same object as approval.stamp, and optionally a stamp_id from request
    """
    if not approval:
        return False
    signoff_stamp_id = getattr(signoff.signet, "stamp_id", None) if signoff else None
    return approval.stamp.id == (
        request_stamp_id or approval.stamp.id
    ) and approval.stamp.id == (signoff_stamp_id or approval.stamp.id)


def get_verify_stamp(approval, kwargs, stamp_id_key="stamp_id"):
    """Helper to return a `verify_stamp` function with correct signature for use with `ApprovalSignoffValidator`"""
    return partial(
        verify_consistent_stamp_id,
        approval,
        request_stamp_id=kwargs.get(stamp_id_key, None),
    )


class BasicUserApprovalActions:
    """
    Concrete implementation that handles simple approvals that don't trigger any follow-up actions

    An extensible base class for extending this simple case that
    implements `SignoffRequestActions` and `ApprovalRequestActions` Protocols
    Facade: delegation & orchestration of signoff_actions_class, and signoff & approval business logic.
    """

    validator_class: SignoffValidator = ApprovalSignoffValidator
    committer_class: SignoffCommitter = ApprovalSignoffCommitter
    signoff_actions_class = BasicUserSignoffActions

    def __init__(
        self,
        user,
        data,
        approval,
        signoff_actions: SignoffRequestActions = None,
        validator: SignoffValidator = None,
        committer: SignoffCommitter = None,
        **kwargs,
    ):
        """
        Define actions available to the given user on the given approval based on request data

        :param user: typically the `request.user`
        :param dict data: a dict-like object, typically with the GET or POST data from the request
        :param AbstractApproval approval: the `Approval` that the signoff/revoke actions are applied to.
        :param SignoffRequestActions signoff_actions: object used to sign approvals, or None to use class default
        :param kwargs: additional parameters passed through to `signoff_actions_class` initializer
        """
        self.user = user
        self.data = data
        self.approval = approval
        self.verify_stamp = get_verify_stamp(approval, kwargs)
        self.validator = validator or self.validator_class(
            user,
            approval,
            verify_signet=get_verify_signet(kwargs),
            verify_stamp=self.verify_stamp,
        )
        committer = committer or self.committer_class(
            user, approval, approve_if_ready=lambda a: self.approve()
        )
        forms = self.signoff_actions_class.form_handler_class(
            self.data, signoff_subject=self.approval
        )
        self.signoff_actions = signoff_actions or self.signoff_actions_class(
            user,
            data,
            form_handler=forms,
            validator=self.validator,
            committer=committer,
            **kwargs,
        )
        self.kwargs = kwargs

    @property
    def signoff(self):
        """Populated by calling sign_ or revoke_signoff"""
        signoff = self.signoff_actions.signoff
        # override the signoff.subject with approval, which might be bound to a higher-level subject,
        #   like an approval_process.  But don't alter the approval type - validation will catch inconsistencies.
        if signoff and (not signoff.subject or signoff.subject == self.approval):
            signoff.subject = self.approval
        return signoff

    @property
    def forms(self):
        """Access to signoff actions form handler object"""
        return self.signoff_actions.forms

    # SignoffRequestActions Protocol - delegate to signoff_actions using approval-level validation and commit logic

    def sign_signoff(self, commit=True):
        """
        Handle request to sign a signoff form for the approval, return True iff signoff succeeded

        :param bool commit: False to validate the form and sign self.signoff, but not commit signet to DB.
        """
        return self.signoff_actions.sign_signoff(commit=commit)

    def revoke_signoff(self, commit=True):
        """
        Handle request to revoke a signoff from this approval, return True iff revoke succeeded

        :param bool commit: False to validate the form, but not actually revoke signet and commit to DB.

        :::{note}
        `commit=False` means "perform validation and return what would have happened, but do not actually revoke"
        There is no way to "revoke" a signoff without committing the change.  Call again to do the actual revoke.
        :::
        """
        return self.signoff_actions.revoke_signoff(commit=commit)

    # ApprovalActions Protocol

    def is_valid_approve_request(self):
        """Return True iff the approval is complete and ready to be approved"""
        return self.verify_stamp() and self.approval.ready_to_approve()

    def approve(self, commit=True):
        """
        Handle request to approve the approval, return True iff it was approved.
        """
        if not self.is_valid_approve_request():
            return False
        self.approval.approve(commit=commit)
        return self.approval.is_approved()

    def is_valid_approval_revoke_request(self):
        """Return True iff user is allowed to revoke the given approval"""
        return self.verify_stamp() and self.approval.can_revoke(self.user)

    def revoke_approval(self, commit=True):
        """
        Handle request to revoke the approval, return True iff approval was revoked.

        :::{note}
        `commit=False` means "perform validation and return what would have happened, but do not actually revoke"
        There is no way to "revoke" approval signoffs without committing the change.  Call again to do the actual revoke.
        :::
        """
        if not self.is_valid_approval_revoke_request():
            return False
        if commit:
            self.approval.revoke(self.user)
        return True


##########################
# APPROVAL PROCESS ACTIONS
##########################


class ApprovalProcessSignoffValidator(ApprovalSignoffValidator):
    """
    Default permissions validation logic for requests to Signoff on an Approval Process

    Implements`SignoffValidator` Protocol
    """

    def __init__(
        self,
        user: User,
        approval: AbstractApproval,
        approval_process: ApprovalsProcess,
        verify_signet: Callable[[AbstractSignoff], bool] = lambda signoff: True,
        verify_stamp: Callable[[AbstractSignoff], bool] = lambda signoff: True,
    ):
        """
        Initialize validator for signoffs on the given approval process.

        :param AbstractApproval approval: that approval for which to validate signoff requests
        :param Callable verify_signet: Optional function to validate request data against a signoff, or None to skip
        :param Callable verify_stamp: Optional function to validate request data, or None to skip

        Adds validation that the stamp being signed is one of the approval_process approval stamps.
        """
        super().__init__(
            user=user,
            approval=approval,
            verify_signet=verify_signet,
            verify_stamp=verify_stamp,
        )
        self.approval_process = approval_process

    def is_valid_signoff_request(self, signoff: AbstractSignoff) -> bool:
        """Return True iff signoff can be signed by user."""
        return super().is_valid_signoff_request(
            signoff
        ) and self.approval_process.user_can_proceed(self.approval, self.user)


def verify_consistent_process_stamp_id(
    approval_process, approval, signoff=None, request_stamp_id=None
):
    """
    Return True iff data relations for approval process, signoff, and request are self-consistent

    Validates that signoff.stamp, if it exists, is same object as approval.stamp, and optionally a stamp_id from request
    """
    return verify_consistent_stamp_id(approval, signoff, request_stamp_id) and (
        approval_process.contains_stamp(request_stamp_id) if request_stamp_id else True
    )


def get_verify_process_stamp(
    approval_process, approval, kwargs, stamp_id_key="stamp_id"
):
    """Helper to return a `verify_stamp` function with correct signature for use with `ApprovalSignoffValidator`"""
    return partial(
        verify_consistent_process_stamp_id,
        approval_process,
        approval,
        request_stamp_id=kwargs.get(stamp_id_key, None),
    )


class ApprovalProcessUserActions(BasicUserApprovalActions):
    """
    Concrete implementation that handles approval actions on an Approval Process model.

    An extensible base class for extending this case that
    implements SignoffRequestActions and ApprovalRequestActions Protocols
    Facade: delegation & orchestration of signoff_actions_class, and signoff & approval business logic and
            a `signoffs.process.ApprovalProcess` object for sequencing and permissions logic
    """

    validator_class: SignoffValidator = ApprovalProcessSignoffValidator

    def __init__(
        self,
        user,
        data,
        approval_process,
        approval=None,
        signoff_actions: SignoffRequestActions = None,
        validator: SignoffValidator = None,
        committer: SignoffCommitter = None,
        **kwargs,
    ):
        """
        Define actions available to the given user on the given approval_process based on request data

        :param user: typically the `request.user`
        :param dict data: a dict-like object, typically with the GET or POST data from the request
        :param approval_process: the `ApprovalProcess` that the signoff/revoke actions are applied to.
        :param approval: the `Approval` to apply signoff/revoke actions to or none to use `next_available_approval`.
        :param signoff_actions: a `SignoffRequestActions` object used to sign approvals, or None to use class default
        :param kwargs: additional parameters passed through to `BasicUserApprovalActions` initializer
        """
        self.approval_process = approval_process
        self.approval = approval or self.approval_process.get_next_available_approval()
        if not self.approval.subject:
            self.approval.subject = self.approval_process.process_model
        validator = validator or self.validator_class(
            user,
            self.approval,
            approval_process,
            verify_signet=get_verify_signet(kwargs),
            verify_stamp=get_verify_process_stamp(
                self.approval_process, self.approval, kwargs
            ),
        )
        super().__init__(
            user,
            data,
            self.approval,
            signoff_actions,
            validator=validator,
            committer=committer,
            **kwargs,
        )

    def is_valid_approve_request(self):
        """Return True iff the approval transition can be completed by user"""
        return self.verify_stamp() and self.approval_process.can_do_approve_transition(
            self.approval, self.user
        )

    def approve(self, commit=True):
        """
        Handle request to make approval transition, return True iff transition was made.

        :::{note}
        `commit=False` likely has no practical value, it is required for Protocol and possible future extensions
        Because transitions are not idempotent.  Call again to do the actual approval.
        :::
        """
        if not self.is_valid_approve_request():
            return False
        return (
            self.approval_process.try_approve_transition(self.approval, self.user)
            if commit
            else True
        )

    def is_valid_approval_revoke_request(self):
        """Return True iff revoke transition for the approval is available to user"""
        return self.verify_stamp() and self.approval_process.can_do_revoke_transition(
            self.approval, self.user
        )

    def revoke_approval(self, commit=True):
        """
        Handle request to make revoke approval transition, return True iff trnasition was made.

        :::{note}
        `commit=False` likely has no practical value, it is required for Protocol and possible future extensions
        Because transitions are not idempotent.  Call again to do the actual revoke.
        :::
        """
        if not self.is_valid_approval_revoke_request():
            return False
        return (
            self.approval_process.try_revoke_transition(self.approval, self.user)
            if commit
            else True
        )



"""
    Core signoffs framework - does not define any concrete models itself.
    Add contrib.signoffs and/or contrib.approvals to add out-of-the-box concrete models.
"""
from django.apps import AppConfig

from signoffs import settings


class SignoffsConfig(AppConfig):
    name = "signoffs"
    default = True
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):
        if settings.SIGNOFFS_AUTODISCOVER_MODULE:
            from django.utils.module_loading import autodiscover_modules

            autodiscover_modules(settings.SIGNOFFS_AUTODISCOVER_MODULE)



"""
Some concrete signoff models for the test app
    Demonstrates how to define custom signoffs and approvals using the contrib.approvals models as a basis.
"""
from django.db import models
from django_fsm import FSMField, transition

from signoffs.approvals import ApprovalSignoff, SimpleApproval
from signoffs.approvals import signing_order as so
from signoffs.models import (
    AbstractRevokedSignet,
    AbstractSignet,
    ApprovalField,
    Signet,
    SignoffField,
    SignoffSet,
    Stamp,
)
from signoffs.process import FsmApprovalsProcess, user_can_revoke_approval
from signoffs.registry import register
from signoffs.signoffs import SignoffRenderer, SimpleSignoff

from . import signoffs

Signet = Signet  # pass-through


# Models for Signoff tests


# Signoffs can be registered in Models to avoid circular imports.
final_report_signoff = SimpleSignoff.register(id="test_app.final_report_signoff")


class Report(models.Model):
    """A model with a related set of signoffs"""

    contents = models.TextField()

    # A "reverse" OneToMany relation backed by the FK defined ono ReportSignet (see below)
    signoffs = SignoffSet(signoffs.report_signoff)

    # A single signoff backed by a OneToOne Field to the signoff's related Signet model
    final_signoff, final_signet = SignoffField(final_report_signoff)


class ReportSignet(AbstractSignet):
    """Persistence layer for Report Signoffs"""

    report = models.ForeignKey(
        Report, on_delete=models.CASCADE, related_name="signatories"
    )


class RevokeReportSignet(AbstractRevokedSignet):
    """Persistence layer for tracking revoked signoffs"""

    signet = models.OneToOneField(
        ReportSignet, on_delete=models.CASCADE, related_name="revoked"
    )


class Vacation(models.Model):
    """A model with both a single Signoff Field and a set of related signoffs"""

    employee = models.CharField(max_length=128)

    # can also create relations using registered signoff id
    employee_signoff, employee_signet = SignoffField("test_app.agree")
    # If signoffs can't be pre-registered (e.g., circular import), you can define the OneToOne field explicitly like:
    # employee_signoff = OneToOneField(Signet, on_delete=models.SET_NULL, null=True, related_name='+')
    # employee_signoff = RelatedSignoff('test_app.agree', employee_signoff)

    signoffset = SignoffSet("test_app.hr_signoff")


class VacationSignet(AbstractSignet):
    """Persistence layer for Vacation Signoffs"""

    vacation = models.ForeignKey(
        Vacation, on_delete=models.CASCADE, related_name="signatories"
    )


# Models for Approval tests


@register(id="test_app.leave_approval")
class LeaveApproval(SimpleApproval):
    """An approval representing a Leave of Absence approval process with a defined SigningOrder"""

    stampModel = Stamp
    label = "Approve Leave of Absence Request"

    employee_signoff_type = ApprovalSignoff.register(
        id="test_app.leave_approval.employee_signoff",
        label="Apply for Leave",
        render=SignoffRenderer(
            form_context=dict(help_text="Employee leave application signoff")
        ),
    )
    hr_signoff_type = ApprovalSignoff.register(
        id="test_app.leave_approval.hr_signoff",
        label="Request Approved by HR",
        render=SignoffRenderer(
            form_context=dict(help_text="HR leave application approval")
        ),
    )
    mngmt_signoff_type = ApprovalSignoff.register(
        id="test_app.leave_approval.mngmt_signoff",
        label="Approve Leave",
        render=SignoffRenderer(
            form_context=dict(help_text="Final leave application approval")
        ),
    )

    signing_order = so.SigningOrder(
        employee_signoff_type, so.AtLeastN(hr_signoff_type, n=1), mngmt_signoff_type
    )


class LeaveRequest(models.Model):
    """A model defining valid set of relations to demonstrate / test ApprovalField"""

    # a One-to-One "forward" approval relation (managed by the Approval business logic) (see approval tests)
    approval = ApprovalField(LeaveApproval)


# Models for FSM Approval Process tests


class Building(models.Model):
    """A model representing a building or construction site"""

    name = models.CharField(max_length=64)
    # ... property_id, address, etc.


# TODO: add permissions to each signoff and to test fixture users

# A suite of signoffs defined to represent the different types of signatures in a Building Permit approval process.
S = ApprovalSignoff
applicant_signoff = S.register(
    "test_app.construction_permit.signoff.applicant",
    label="Apply for Permit",
    render=SignoffRenderer(
        form_context=dict(help_text="Building permit application - applicant signoff")
    ),
)
planning_signoff = S.register(
    "test_app.construction_permit.signoff.planning",
    label="Application Meets By-lawas",
    render=SignoffRenderer(
        form_context=dict(help_text="Building permit application - planning signoff")
    ),
)
permit_approval = S.register(
    "test_app.construction_permit.signoff.permit", label="Approve Building Permit"
)
electrical_signoff = S.register(
    "test_app.construction_permit.signoff.electrical",
    label="Application / Installation Meets Electrical Code",
)
plumbing_signoff = S.register(
    "test_app.construction_permit.signoff.plumbing",
    label="Application / Installation Meets Plumbing Code",
)
inspection_approval = S.register(
    "test_app.construction_permit.signoff.inspection",
    label="Construction Inspected",
    render=SignoffRenderer(
        form_context=dict(
            help_text="Construction Inspection - construction meets applicable standards"
        )
    ),
)


class AbstractBuildingPermitApproval(SimpleApproval):
    """Abstract base class for Building Permit Approvals defining the business logic for signing the approval"""

    def sign_approval(self, user, last=False):
        """Sign the first or last of the next available signoffs for given user"""
        next = self.next_signoffs(for_user=user)
        if next and not self.is_approved() and not self.has_signed(user):
            index = -1 if last else 0
            next[index].sign(user)


@register(id="test_app.building_permit.permit_application")
class BuildingPermitApplication(AbstractBuildingPermitApproval):
    """Building Permit Application Approval Type"""

    signing_order = so.SigningOrder(applicant_signoff)


@register(id="test_app.building_permit.permit_approval")
class BuildingPermitApproval(AbstractBuildingPermitApproval):
    """Building Permit Application Approval Type"""

    signing_order = so.SigningOrder(
        planning_signoff, electrical_signoff, permit_approval
    )


@register(id="test_app.building_permit.interim_inspection_approval")
class InterimInspectionApproval(AbstractBuildingPermitApproval):
    """Building Permit Interim Inspection Approval Type"""

    signing_order = so.SigningOrder(
        so.InParallel(
            so.OneOrMore(electrical_signoff),
            so.OneOrMore(plumbing_signoff),
        ),
        inspection_approval,
    )


@register(id="test_app.building_permit.final_inspection_approval")
class FinalInspectionApproval(AbstractBuildingPermitApproval):
    """Building Permit Final Inspection Approval Type"""

    signing_order = so.SigningOrder(
        so.InParallel(
            so.Optional(electrical_signoff),
            so.Optional(plumbing_signoff),
        ),
        inspection_approval,
    )


class ConstructionPermittingProcess(models.Model):
    """Defines a 5-stage Approval Process for applying, permitting, and inspecting building permits"""

    class States(models.TextChoices):
        INITIATED = "Initiated"
        APPLIED = "Applied"
        PERMITTED = "Permitted"
        INSPECTED = "Inspected"
        APPROVED = "Approved"

    # Model state + relation to the building permit is for
    state = FSMField(choices=States.choices, default=States.INITIATED)
    building = models.ForeignKey(
        Building, on_delete=models.CASCADE, related_name="permits"
    )

    # Four OnetoOne fields defining the approvals needed in the process
    apply, apply_stamp = ApprovalField(BuildingPermitApplication)
    permit, permit_stamp = ApprovalField(BuildingPermitApproval)
    interim_inspection, interim_inspection_stamp = ApprovalField(
        InterimInspectionApproval
    )
    final_inspection, final_inspection_stamp = ApprovalField(FinalInspectionApproval)

    process = FsmApprovalsProcess()

    # Approval / FSM transitions defining state transitions and their side effects.

    # transitions can be decorated the verbose (flexible) way... (note: decorator order is imporant here!)

    @process.register_approve_transition(apply)
    @transition(field=state, source=States.INITIATED, target=States.APPLIED)
    @process.do_approval
    def applied(self, approval):
        print("Applied", self.state, approval)

    @process.register_revoke_transition(apply)
    @transition(
        field=state,
        source=States.APPLIED,
        target=States.INITIATED,
        permission=user_can_revoke_approval(apply),
    )
    @process.do_revoke
    def rejected(self, approval):
        print("Rejected Application", self.state, approval)

    # ... or the convenient way...

    @process.approval_transition(
        permit, field=state, source=States.APPLIED, target=States.PERMITTED
    )
    def permitted(self, approval):
        print("Permitted", self.state, approval)

    @process.approval_transition(
        interim_inspection,
        field=state,
        source=States.PERMITTED,
        target=States.INSPECTED,
    )
    def inspected(self, approval):
        print("Interim Inspected", self.state, approval)

    @process.approval_transition(
        final_inspection, field=state, source=States.INSPECTED, target=States.APPROVED
    )
    def authorized(self, approval):
        print("Final Inspection Approved", self.state, approval)



"""
App-dependent tests for signoff forms - requires models that can be saved, etc.
"""
from django.core import exceptions
from django.test import TestCase

from signoffs.core.tests import fixtures
from signoffs.forms import signoff_form_factory
from tests.test_app import models, signoffs

consent_signoff = signoffs.consent_signoff
SignoffForm = signoff_form_factory(signoff_type=consent_signoff)

ReportSignoffForm = signoff_form_factory(signoff_type=signoffs.report_signoff)


class SignoffFormWithUserTests(TestCase):
    def get_form(self, data=None, **kwargs):
        data = data or dict(signed_off="True", signoff_id=consent_signoff.id)
        return SignoffForm(data=data, **kwargs)

    def test_save(self):
        u = fixtures.get_user(perms=("can_sign",))
        bf = self.get_form()
        self.assertTrue(bf.is_valid())
        v = bf.sign(user=u)
        self.assertTrue(v.is_signed)
        self.assertEqual(v.id, consent_signoff.id)
        self.assertEqual(v.signet.user, u)
        self.assertEqual(v.signet.sigil, u.get_full_name())

    def test_invalid_save(self):
        bf = self.get_form()
        self.assertTrue(
            bf.is_valid()
        )  # form is valid even if user doesn't have permission to save it.
        with self.assertRaises(exceptions.PermissionDenied):
            bf.sign(user=None)

    def test_save_no_perm(self):
        u = fixtures.get_user(username="NoCanDoBoo")
        bf = self.get_form()
        self.assertTrue(bf.is_valid())
        with self.assertRaises(exceptions.PermissionDenied):
            bf.sign(user=None)  # no user
        with self.assertRaises(exceptions.PermissionDenied):
            bf.sign(user=u)  # unpermitted user


class SignoffFormWithRelationTests(TestCase):
    def get_form(self, **kwargs):
        data = dict(signed_off="True", signoff_id=signoffs.report_signoff.id, **kwargs)
        return ReportSignoffForm(data=data)

    def test_save_with_relation(self):
        u = fixtures.get_user(perms=("can_review",))
        r = models.Report.objects.create(contents="Awesome report contents.")
        bf = self.get_form(report=r)
        self.assertTrue(bf.is_valid())
        v = bf.sign(user=u)
        signet = models.ReportSignet.objects.select_related("report").get(
            pk=v.signet.pk
        )
        self.assertEqual(v.signet.report, r)
        report = models.Report.objects.prefetch_related("signatories").get(pk=r.pk)
        self.assertEqual(report.signatories.count(), 1)
        self.assertEqual(report.signatories.first(), signet)



"""
    An Approval manages logic for collecting and sequencing one or more Signoffs.

    `Approval Types` are registered subclasses of AbstractApproval
     - they define the behaviour, sequencing logic, and state transition logic for an Approval instance.

    Persistence layer for `Approval` state is provided by a `Stamp` model (think "Stamp of Approval" or TimeStamp)
     - one concrete `Stamp` model can back any number of Approval Types
     - an `Approval` provides the business and presentation logic for a `Stamp` instance.
"""
from typing import Callable, Optional, Type, Union

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.db import transaction
from django.utils.text import slugify

from signoffs.core import models, utils
from signoffs.core.models import managers
from signoffs.core.renderers import ApprovalRenderer
from signoffs.core.signing_order import SigningOrder
from signoffs.core.status import ApprovalStatus
from signoffs.core.urls import ApprovalUrlsManager

# type name shorts
opt_str = Union[bool, Optional[str]]
stamp_type = Union[str, Type[models.AbstractApprovalStamp]]


# The business logic for revoking an approval may be dependent on context that shouldn't be encoded in an Approval
# Application logic may also need a way to revoke approvals without triggering Approval permission or signal logic.
# revoke_approval provides the default implementation for basic revoke business logic.
# Implementors can override the behaviour of Approval.revoke() by either overriding the class method or by injecting
#      a function with the same signature as the default implementation provided here...


def revoke_approval(approval, user, reason=""):
    """
    Force revoke the given approval for user regardless of permissions or approval state!

    Default implementation revokes ALL related signets on behalf of the user
      - a user with permission to revoke an approval must have permission to revoke all signoffs within!
    """
    with transaction.atomic():
        # First mark approval as no longer approved, b/c signoffs can't be revoked from approved approval
        approval.stamp.approved = False
        # Revoke all signoffs in reverse order
        for signoff in reversed(approval.signatories.signoffs()):
            signoff.revoke(user=user, reason=reason)

        approval.save()


class DefaultApprovalBusinessLogic:
    """Defines the business logic for Approving and Revoking an Approval instance"""

    # Process / permissions to revoke an approval of this Type: False for irrevocable;  None (falsy) for unrestricted
    revoke_perm: opt_str = ""  # e.g. 'approvals.delete_stamp'
    revoke_method: Callable = revoke_approval  # revoke approval algorithm

    def __init__(self, revoke_perm=None, revoke_method=None):
        """Override default actions, or leave parameter None to use class default"""
        self.revoke_perm = revoke_perm if revoke_perm is not None else self.revoke_perm
        self.revoke_method = (
            revoke_method or type(self).revoke_method
        )  # don't bind revoke_method to self here

    # Approve Actions / Rules

    def is_signable(self, approval, by_user=None):
        """
        Return True iff this approval is in a state it could be signed by the given user

        :::{important}
         - this is approval-level logic only -- keep signoff-level rules in signoff.can_sign
         - does not determine if there is a signoff available to be signed, only about the state of this approval!
         - use can_sign to determine if there are any actual signoffs available to the user to be signed.
        :::
        """
        return not approval.is_approved()

    def can_sign(self, approval, user, signoff=None):
        """
        Return True iff the given user can sign given signoff on this approval,
            or any of the next signoffs in its signing order

        If a `Signoff` instance is provided, check that the user can sign this specific signoff.
        """
        avaialable = approval.next_signoffs(
            for_user=user
        )  # assert: all(s.can_sign(user) for s in available)
        return self.is_signable(approval, by_user=user) and (
            any(s.matches(signoff) for s in avaialable)
            if signoff
            else len(avaialable) > 0
        )

    def ready_to_approve(self, approval):
        """Return True iff the approval's signing order is complete and ready to be approved"""
        return not approval.is_approved() and approval.is_complete()

    def approve_if_ready(self, approval, commit=True):
        """Approve and save the approval is it meets all ready conditions"""
        if self.ready_to_approve(approval):
            self.approve(approval, commit)

    def approve(self, approval, commit=True, **kwargs):
        """
        Approve the approval and save it's Stamp.

        No permissions or signoff completion logic involved here - just force into approved state!
         - Prefer to use `approve_if_ready` to enforce business rules.
        `kwargs` passed directly to save() - use commit=False to approve without saving
        """
        approval.stamp.approve()
        if commit:
            approval.save(**kwargs)

    # Revoke Actions / Rules

    def is_revokable(self, approval, by_user=None):
        """
        Return True iff this approval is in a state it could be revoked by the given user

        :::{important}
         - this is approval-level logic only -- keep signoff-level rules in signoff.can_revoke
         - does not determine if there is a signoff available to be revoked, only about the state of this approval!
         - use can_revoke to determine if the approval is actually available to the user to be revoked.
        :::
        """
        return approval.is_approved()

    def is_permitted_revoker(self, approval_type, user):
        """return True iff user has permission to revoke approvals of given Type"""
        revoke_perm = self.revoke_perm
        return (
            False
            if self.revoke_perm is False
            else user.has_perm(revoke_perm)
            if revoke_perm
            else True
        )

    def can_revoke(self, approval, user):
        """Return True iff the approval can be revoked by given user"""
        # Note: assumes a user with permission to revoke an approval would also have permission to revoke all signoffs within.
        return self.is_revokable(approval, user) and self.is_permitted_revoker(
            type(approval), user
        )

    def revoke_if_permitted(self, approval, user, reason=""):
        """
        Revoke and save the approval if it meets all conditions for revocation.

        :raises PermissionDenied: if not
        """
        if not self.can_revoke(approval, user):
            raise PermissionDenied(
                f"User {user} does not have permission to revoke approval {self}"
            )

        return self.revoke(approval, user, reason)

    def revoke(self, approval, user, reason=""):
        """
        Revoke the approval and save it's Stamp.

        No permissions or completion logic involved here - just force into revoked state!
        Prefer to use `revoke_if_permitted` to enforce business rules.
        """
        return self.revoke_method(approval, user, reason)

    def can_revoke_signoff(self, approval, signoff, user):
        """
        Return True iff the given signoff can be revoked from this approval by given user

        :::{note}
        Default logic restricts revoke to the last signoff collected on unapproved approvals.
        Think carefully before overriding this restriction - users sign in-order and that often has meaning,
        even in cases where signoffs are collected purely "in-parallel".
        """
        return (
            not approval.is_approved()
            and signoff.can_revoke(user)
            and signoff == approval.signoffs.latest()
        )


class ApprovalLogic(DefaultApprovalBusinessLogic):
    """Public API: Alias for `DefaultApprovalBusinessLogic"""


class AbstractApproval:
    """
    Defines the semantics for Approving something using a sequence of one or more `Signoffs`

    An Approval Type (subclass of `AbstractApproval`) defines the business and presentation logic
    for a specific type of approval.
    The state of an `Approval` instance is persisted by an `ApprovalStamp` and its related `Signets`.

    Approval Types are pure-code objects, not stored in DB, as they define application logic, not application data.
    An Approval Type defines:
      - how the `Stamp` is labelled and rendered,
      - what sequecne of signoffs is required to complete it;
      - what permission is required to revoke it,
     - define and register new Approval Types with factory method:: `BaseApproval.register(...)`

    Default meta-data & services can be overridden by subclasses or passed to `.register()` factory
    Approval Types are registered in the `signoffs.registry.approvals` where they can be retrieved by `id`
    :::{caution}
    `Stamp` records are stored in DB with a reference to `Approval.id`
    Be cautions not to change or delete id's that are in-use.
    :::

    Use a `SigningOrder` to define the sequence of Signgoffs required.
     - default signing order is sequential, ordered alphabetically,
       based on `SignoffFields` / attribute defined on the `Approval`
    :::{tip}
    - most Approvals will override `signing_order`
    - use ordering API on Approval instance rather than accessing `signing_order` directly!
    :::
    """

    # id must be unique per type class, but human-legible / meaningful - dotted path recommended e.g. 'myapp.approval'
    id: str = "approval.abstract"
    """unique identifier for type - used like FK, don't mess with these!"""

    # stampModel is required - every Approval Type must supply a concrete Stamp model to provide persistence layer
    stampModel: stamp_type = None  # concrete Model or 'app.model' string - REQUIRED

    # Manager for the entire collection of signoffs related to an Approval instance
    signoffsManager: type = managers.StampSignoffsManager  # injectable Manager class
    # Optional Signing Order Manager drives ordering API to determine "next" signoff available to a given user.
    signing_order: SigningOrder = None  # sequencing logic for approval's signoffs

    # Approval business logic, actions, and permissions
    logic: ApprovalLogic = ApprovalLogic()

    # Accessor for approval status info
    status: ApprovalStatus = ApprovalStatus()

    # Define visual representation for approvals of this Type. Label is a rendering detail, but common override.
    label: str = ""  # Label for the Approval empty string for no label
    render: ApprovalRenderer = ApprovalRenderer()  # presentation logic service
    urls: ApprovalUrlsManager = ApprovalUrlsManager()  # service to provide endpoints

    # Registration for Approval Types (aka subclass factory)

    @classmethod
    def register(cls, id, **kwargs):
        """
        Create, register, and return a new subclass of cls with overrides for given kwargs attributes

        Standard mechanism to define new Approval Types, typically in `my_app/models.py` or `my_app/approvals.py`
        Usage:
        ```
        MyApproval = AbstractApproval.register('my_appproval_type', label='Approve it!', ...)
        ```
        """
        from signoffs import registry

        class_name = utils.id_to_camel(id)
        kwargs["id"] = id
        approval_type = type(class_name, (cls,), kwargs)
        registry.approvals.register(approval_type)
        return approval_type

    @classmethod
    def validate(cls):
        """Run any class validation that must pass before class can be registered.  Invoked by registry."""
        if cls.stampModel is None:
            raise ImproperlyConfigured(
                f"Approval Type {cls.id} must specify a Stamp Model."
            )
        return True

    # Approval Type accessors

    @classmethod
    def get_stampModel(cls):
        """Always use this accessor as the stampModel attribute may be an "app.Model" label"""
        if not cls.stampModel:
            raise ImproperlyConfigured(
                f"No Stamp Model associated with Approval {cls}."
            )
        if isinstance(cls.stampModel, str):
            cls.stampModel = apps.get_model(cls.stampModel)
        return cls.stampModel

    @classmethod
    def get_stamp_queryset(cls, prefetch=("signatories",)):
        """Return a base (unfiltered) queryset of ALL Stamps for this Approval Type"""
        return (
            cls.get_stampModel()
            .objects.filter(approval_id=cls.id)
            .prefetch_related(*prefetch)
            .all()
        )

    # Approval Type behaviours

    @classmethod
    def create(cls, **kwargs):
        """Create and return an approval backed by a new Stamp instance"""
        approval = cls(**kwargs)
        approval.save()
        return approval

    # Approval instance behaviours

    def __init__(self, stamp=None, subject=None, **kwargs):
        """
        Construct an Approval instance backed by the given stamp or an instance of cls.stampModel(**kwargs)
        subject is optional: the object this approval is meant to approve - set by ApprovalField but otherwise unused.
        """
        self.stamp = stamp or self.get_new_stamp(**kwargs)
        self._subject = subject
        if not self.stamp.approval_id == self.id:
            raise ImproperlyConfigured(
                f"Approval Type {self} does not match Stamp Model {self.stamp.approval_id}."
            )

    @property
    def subject(self):
        """
        The object being approved, if provided.
        Sub-classes with stamp FK relations may want to override this to access the stamp related object.
        subject is set by model Fields for convenient access to owner obj, but value is not used by any core logic.
        """
        return self._subject

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def slug(self):
        """A slugified version of the signoff id, for places where a unique identifier slug is required"""
        return slugify(self.id)

    @property
    def stamp_model(self):
        """return the signoff model for this type"""
        return self.get_stampModel()

    def get_new_stamp(self, **kwargs):
        """Get a new, unsaved stamp instance for this approval type"""
        Stamp = self.stamp_model
        return Stamp(approval_id=self.id, **kwargs)

    def __str__(self):
        return str(self.stamp)

    def __eq__(self, other):
        return (
            isinstance(other, AbstractApproval)
            and self.id == other.id
            and self.stamp == other.stamp
        )

    def __contains__(self, item):
        """Return True iff this approval's signatories contains the item: signoff id, type, or user"""
        # assumes item is either a user or otherwise a signoff id, type, or instance.
        return (
            self.has_signed(item)
            if hasattr(item, "username")
            else self.has_signoff(item)
        )

    # Signoff Manager accessors

    @property
    def signoffs(self):
        """
        Return an ApprovalSignoffsManager for access to this approval's signoff set
        Default implementation returns manager for signoffs backed by stamp's signet_set manager.
        Concrete Approval Types may inject a custom signoffsManager for custom set of signoffs.
        """
        return self.signoffsManager(self.stamp, subject=self)

    def is_signed(self):
        """Return True iff this approval has at least one signoff"""
        return self.signoffs.exists()

    def has_signed(self, user):
        """Return True iff given user is a signatory on this approval's set of signoffs"""
        return any(s.user == user for s in self.signatories.all())

    def has_signoff(self, signoff_id_or_type):
        """Return True iff this approval already has a signoff of the given signoff_type"""
        from signoffs import registry

        try:
            signoff_type = registry.get_signoff_type(signoff_id_or_type)
        except ImproperlyConfigured:
            return False
        return any(s.signoff_id == signoff_type.id for s in self.signatories.all())

    # Approval Business Logic Delegation

    def can_sign(self, user, signoff=None):
        """
        return True iff the given user can sign given signoff on this approval,
            or any of the next signoffs in its signing order
        """
        return self.logic.can_sign(self, user, signoff)

    def ready_to_approve(self):
        """Return True iff this approval's signing order is complete and ready to be approved"""
        return self.logic.ready_to_approve(self)

    def approve_if_ready(self, commit=True):
        """Approve and save this approval is it meets all ready conditions"""
        return self.logic.approve_if_ready(self, commit)

    def approve(self, commit=True, **kwargs):
        """
        Approve this stamp. No permissions or signoff logic involved here - just force into approved state!
        Raises PermissionDenied if self.is_approved() -- can't approval an approved approval :-P
        kwargs passed directly to save() - use commit=False to approve without saving
        """
        return self.logic.approve(self, commit=commit, **kwargs)

    def is_revokable(self):
        """Return True iff this approval is in a state it could be revoked"""
        return self.logic.is_revokable(self)

    @classmethod
    def is_permitted_revoker(cls, user):
        """Return True iff user has permission to revoke approvals of this type"""
        return cls.logic.is_permitted_revoker(cls, user)

    def can_revoke(self, user):
        """Return True iff this approval can be revoked by given user"""
        return self.logic.can_revoke(self, user)

    def revoke_if_permitted(self, user, **kwargs):
        """Revoke and save the approval is it meets all conditions for revocation"""
        return self.logic.revoke_if_permitted(self, user, **kwargs)

    def revoke(self, user, **kwargs):
        """Revoke this approval regardless of permissions or approval state - careful!"""
        return self.logic.revoke(self, user, **kwargs)

    def can_revoke_signoff(self, signoff, user):
        """Return True iff the given signoff can be revoked from this approval by given user"""
        return self.logic.can_revoke_signoff(self, signoff, user)

    # Stamp Delegation

    @property
    def signatories(self):
        """
        Return queryset of Signets representing the signatories on this approval

        Default implementation simply returns Stamp's "reverse" signet_set manager.
        Concrete Approval Types can override to provide any sensible qs of signets.
        """
        return self.stamp.signatories

    @property
    def timestamp(self):
        """Return the timestamp approval was granted, None otherwise"""
        return self.stamp.timestamp if self.is_approved() else None

    def has_signatories(self):
        """Return True iff this approval has any signatories"""
        return len(self.signatories.all()) > 0

    def is_approved(self):
        """Return True iff this Approval is in an approved state"""
        return self.stamp.is_approved()

    def save(self, *args, **kwargs):
        """Attempt to save the Stamp of Approval, with the provided given associated data"""
        self.stamp.save(*args, **kwargs)
        return self

    @classmethod
    def has_object_relation(cls):
        return cls.get_stampModel().has_object_relation()

    # SigningOrder delegates - signing_order on approval instance is a SigningOrderManager
    # Assumptions:
    #  - signoffs have a "stamp" relation (FK to Stamp) -  SigningOrder requires a "reverse" set accessor
    #   (could loosen this assumption by making name of that relation configurable with "stamp" as default

    def is_complete(self):
        """
        Is this approval process complete and ready to be approved?
        Default implementation returns False if no signing order, True if the signing order is complete.
        Concrete Approval Types can override this method to customize conditions under which this approval is complete.
        """
        return bool(self.signing_order and self.signing_order.is_complete())

    def next_signoff_types(self, for_user=None):
        """
        Return list of next signoff type(s) (Signoff Type) required in this approval process.

        Default impl returns next signoffs from the approval's signing order or [] if no signing order is available.
        Concrete Approval Types can override this with custom business logic to provide signing order automation.
        """
        signoff_types = self.signing_order.next_signoffs() if self.signing_order else []
        return [
            signoff
            for signoff in signoff_types
            if (for_user is None or signoff.is_permitted_signer(for_user))
        ]

    def next_signoffs(self, for_user=None):
        """
        Return list of next signoff instance(s) required in this approval process.

        :returns: list[AbstractSignoff] where all(s.can_sign(for_user) for s in list)

        If a user object is supplied, filter out instances not available to that user.
        Most applications will define custom business logic for ordering signoffs, restricting duplicate signs, etc.
            - ideally, use ApprovalLogic and SigningOrder to handle these, but this gives total control!
        """
        if not self.logic.is_signable(self, for_user):
            return []
        signoffs = (
            signoff(stamp=self.stamp, subject=self, user=for_user)
            for signoff in self.next_signoff_types(for_user)
        )
        return [s for s in signoffs if for_user is None or s.can_sign(for_user)]

    def get_next_signoff(self, for_user=None):
        """
        Return the next available signoffs for given user, or None

        Again, ideally define ApprovalLogic or SigningOrder rather than overriding behaviour here.
        """
        next = self.next_signoffs(for_user=for_user)
        return next[0] if next else None


class BaseApproval(AbstractApproval):
    """
    A base Approval Type to be used as base class or to register concrete Approval Types
    Concrete Types will require a concrete Stamp Model to back the approval.
    """

    id = "signoffs.base-approval"
    stampModel = None


__all__ = [
    "revoke_approval",
    "AbstractApproval",
    "BaseApproval",
    "DefaultApprovalBusinessLogic",
    "ApprovalLogic",
]
